from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings
from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestCoilPlayer(MpfTestCase):

    def get_config_file(self):
        return 'coil_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/coil_player/'

    def get_platform(self):
        return "smart_virtual"

    def test_express_config(self):
        # coil with allow_enable set
        coil = self.machine.coils['coil_1']
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()

        self.machine.events.post('event1')
        self.advance_time_and_run()

        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=10))
        assert not coil.hw_driver.enable.called

        # coil without allow_enable
        coil = self.machine.coils['coil_2']
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()

        self.machine.events.post('event6')
        self.advance_time_and_run()

        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=10))
        assert not coil.hw_driver.enable.called

        coil = self.machine.coils['coil_3']
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.disable = MagicMock()

        self.post_event("event7")
        self.advance_time_and_run()
        coil.hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=10), HoldSettings(power=0.5))
        assert not coil.hw_driver.disable.called
        assert not coil.hw_driver.pulse.called
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.disable = MagicMock()

        self.post_event("event8")
        self.advance_time_and_run()
        coil.hw_driver.disable.assert_called_with()
        assert not coil.hw_driver.enable.called
        assert not coil.hw_driver.pulse.called
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.disable = MagicMock()

        self.post_event("event9")
        self.advance_time_and_run()
        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=30))
        assert not coil.hw_driver.disable.called
        assert not coil.hw_driver.enable.called

    def test_pulse(self):
        coil = self.machine.coils['coil_1']
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()

        # coil without allow_enable
        coil2 = self.machine.coils['coil_2']
        coil2.hw_driver.pulse = MagicMock()
        coil2.hw_driver.enable = MagicMock()

        self.machine.events.post('event2')
        self.advance_time_and_run()

        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=10))
        assert not coil.hw_driver.enable.called

        coil2.hw_driver.pulse.assert_called_with(PulseSettings(power=0.5, duration=10))
        assert not coil2.hw_driver.enable.called

        # post same event again
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()
        coil2.hw_driver.pulse = MagicMock()
        coil2.hw_driver.enable = MagicMock()
        self.machine.events.post('event2')
        self.advance_time_and_run()

        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=10))
        assert not coil.hw_driver.enable.called

        coil2.hw_driver.pulse.assert_called_with(PulseSettings(power=0.5, duration=10))
        assert not coil2.hw_driver.enable.called

    def test_pulse_with_attributes(self):
        coil = self.machine.coils['coil_1']
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.enable = MagicMock()

        self.machine.events.post('event3')
        self.advance_time_and_run()

        coil.hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=49))
        assert not coil.hw_driver.enable.called

    def test_enable_and_disable(self):
        coil = self.machine.coils['coil_1']
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.disable = MagicMock()

        self.machine.events.post('event4')
        self.advance_time_and_run()

        coil.hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=10), HoldSettings(power=1.0))
        assert not coil.hw_driver.pulse.called

        self.machine.events.post('event5')
        self.advance_time_and_run()

        coil.hw_driver.disable.assert_called_with()

        # same again but use on and off instead of enable and disable
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.pulse = MagicMock()
        coil.hw_driver.disable = MagicMock()

        self.machine.events.post('event10')
        self.advance_time_and_run()

        coil.hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=10), HoldSettings(power=1.0))
        assert not coil.hw_driver.pulse.called

        self.machine.events.post('event11')
        self.advance_time_and_run()

        coil.hw_driver.disable.assert_called_with()

    def test_coil_player_in_mode(self):
        coil = self.machine.coils['coil_3']
        coil.hw_driver.enable = MagicMock()
        coil.hw_driver.disable = MagicMock()

        # start mode
        self.post_event("start_mode1", 1)
        self.assertFalse(coil.hw_driver.enable.called)
        self.assertFalse(coil.hw_driver.disable.called)

        # enable coil
        self.post_event("event1_mode", 1)
        coil.hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=10), HoldSettings(power=1.0))
        coil.hw_driver.enable = MagicMock()
        self.assertFalse(coil.hw_driver.disable.called)

        # on mode stop the coil player should disable the coil
        self.post_event("stop_mode1", 1)
        self.assertFalse(coil.hw_driver.enable.called)
        coil.hw_driver.disable.assert_called_with()

    def test_max_wait_ms(self):
        coil = self.machine.coils['coil_1']
        self.post_event("pulse_1_100")
        self.advance_time_and_run(.05)
        self.assertEqual("pulsed_100", coil.hw_driver.state)
        self.post_event("pulse_1_50_max_wait_ms")
        # still the same
        self.assertEqual("pulsed_100", coil.hw_driver.state)
        # after pulse end
        self.advance_time_and_run(.06)
        self.assertEqual("pulsed_50", coil.hw_driver.state)



# pylint: disable-msg=too-many-lines
"""Config Migrator for v4."""
import os
import re
from copy import deepcopy

from typing import Dict
from typing import Tuple

from ruamel import yaml
from ruamel.yaml.comments import CommentedMap, CommentedSeq

from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip, MpfRoundTripLoader
from mpf.core.utility_functions import Util
from mpf.migrator.migrator import VersionMigrator
from mpf.core.rgb_color import NAMED_RGB_COLORS, RGBColor


class V4Migrator(VersionMigrator):

    """Migrate config to v4."""

    config_version = 4

    # These sections are in the order they're processed

    renames = '''
    - old: sound_system|initial_volume
      new: master_volume
    - old: tocks_per_sec
      new: speed
    - old: fonts
      new: widget_styles
    - old: movies
      new: videos
    - old: lights_when_disabled
      new: show_when_disable
    - old: num_repeats
      new: loops
    - old: repeat
      new: loops
    - old: loop
      new: loop
    - old: game|allow start with loose balls
      new: allow_start_with_loose_balls
    '''

    moves = '''
    - old: window|frame
      new: window|borderless
    - old: window|quit_on_close
      new: window|exit_on_escape
    '''

    deprecations = '''
    - timing
    - plugins
    - sound_system|volume_steps
    - sound_system|stream
    - window|elements|__list__|pixel_spacing
    - window|fps
    - dmd
    - displays|dmd|physical
    - displays|dmd|type

    # everything from here down is old than v3, but I saw them in some configs
    # so figured we can get rid of them now too
    - machine_flow
    - machineflow
    '''

    additions = '''
    sound_system:
      enabled: True
    '''

    slides = dict()             # type: Dict[str, int]
    displays = dict()           # type: Dict[str, Tuple[int, int]]
    default_display = None      # type: str
    color_dmd = False
    WIDTH = 800
    HEIGHT = 600
    MAIN_CONFIG_FILE = 'config.yaml'

    @classmethod
    def _get_slide_name(cls, display):
        if display not in cls.slides:
            cls.slides[display] = 0

        cls.slides[display] += 1

        if display:
            return '{}_slide_{}'.format(display, cls.slides[display])

        return 'slide_{}'.format(cls.slides[display])

    @classmethod
    def _add_display(cls, name, w, h):
        cls.log.debug("Detected display '%s' (%sx%s)", name, w, h)
        cls.displays[name] = (w, h)

    def _do_custom(self):
        self._set_dmd_type()
        self._migrate_window()
        self._create_display_from_dmd()
        self._migrate_physical_dmd()
        self._set_default_display()
        self._migrate_slide_player()
        self._create_window_slide()
        self._migrate_sound_system()
        self._migrate_fonts()
        self._migrate_asset_defaults()
        self._migrate_animation_assets()
        self._migrate_show_player()
        self._migrate_shots()
        self._migrate_light_player()
        self._migrate_shot_profiles()
        self._migrate_light_scripts()
        self._migrate_mode_timers()
        self._migrate_assets('images')
        self._migrate_assets('videos')
        self._migrate_assets('sounds')
        self._migrate_switches()
        self._migrate_sound_player()
        self._migrate_logic_blocks()

    def _migrate_mode_timers(self):
        if 'timers' in self.fc:
            for timer_name in self.fc['timers']:
                if "start_paused" in self.fc['timers'][timer_name]:
                    if not self.fc['timers'][timer_name]['start_paused']:
                        self.fc['timers'][timer_name]["start_running"] = True
                    elif ("start_running" not in self.fc['timers'][timer_name] or
                            not self.fc['timers'][timer_name]["start_running"]):
                        self.fc['timers'][timer_name]["start_running"] = False
                    else:
                        raise ValueError("Both start_paused and start_running are true. This is impossible")

                    YamlRoundtrip.del_key_with_comments(self.fc['timers'][timer_name], 'start_paused', self.log)

    def _migrate_window(self):
        # Create a display from the window
        if 'window' in self.fc:
            self.log.debug("Converting window: section")
            if 'displays' not in self.fc:
                self.fc['displays'] = CommentedMap()
            self.fc['displays']['window'] = CommentedMap()

            self.fc['displays']['window']['height'] = self.fc['window'].get(
                'height', 600)
            self.fc['displays']['window']['width'] = self.fc['window'].get(
                'width', 800)
            self._add_display('window', self.fc['displays']['window']['width'],
                              self.fc['displays']['window']['height'])
            V4Migrator.default_display = 'window'

            try:  # old setting was 'frame', so we need to flip it
                self.fc['window']['borderless'] = not self.fc['window']['borderless']
            except KeyError:
                pass

    def _create_display_from_dmd(self):
        if 'dmd' in self.fc:
            self.log.debug("Converting dmd: to displays:dmd:")
            if 'displays' not in self.fc:
                self.log.debug("Creating 'displays:' section")
                self.fc['displays'] = CommentedMap()

            V4Migrator.default_display = 'dmd'

            self.log.debug("Creating 'displays:dmd: section")
            self.fc['displays']['dmd'] = CommentedMap()

            self.fc['displays']['dmd'].update(self.fc['dmd'])
            self._add_display('dmd', self.fc['dmd']['width'],
                              self.fc['dmd']['height'])

    def _migrate_physical_dmd(self):
        if ('dmd' in self.fc and 'physical' in self.fc['dmd'] and
                self.fc['dmd']['physical']):

            self.log.debug("Converting physical dmd: settings")

            YamlRoundtrip.del_key_with_comments(self.fc['dmd'], 'physical',
                                                self.log)
            YamlRoundtrip.del_key_with_comments(self.fc['dmd'], 'fps',
                                                self.log)

            if 'type' in self.fc['dmd'] and self.fc['dmd']['type'] == 'color':
                # physical color DMD
                YamlRoundtrip.del_key_with_comments(self.fc['dmd'], 'type',
                                                    self.log)
                YamlRoundtrip.rename_key('dmd', 'physical_rgb_dmd', self.fc,
                                         self.log)

            else:  # physical mono DMD
                YamlRoundtrip.del_key_with_comments(self.fc['dmd'], 'type',
                                                    self.log)

                YamlRoundtrip.rename_key('dmd', 'physical_dmd', self.fc,
                                         self.log)

            YamlRoundtrip.del_key_with_comments(self.fc['displays']['dmd'],
                                                'physical', self.log)
            YamlRoundtrip.del_key_with_comments(self.fc['displays']['dmd'],
                                                'shades', self.log)
            YamlRoundtrip.del_key_with_comments(self.fc['displays']['dmd'],
                                                'fps', self.log)

    def _set_dmd_type(self):
        if 'dmd' in self.fc and 'type' in self.fc['dmd']:
            if self.fc['dmd']['type'].lower() == 'color':
                V4Migrator.color_dmd = True
                self.log.debug("Detected settings for color DMD")

    def _set_default_display(self):
        if 'displays' not in self.fc or len(self.fc['displays']) == 1:
            return

        if V4Migrator.default_display in self.fc['displays']:
            self.fc['displays'][V4Migrator.default_display]['default'] = True

    def _migrate_slide_player(self):
        if 'slide_player' not in self.fc:
            return

        self.log.debug("Converting slide_player: entries to slides:")

        self.fc['slides'] = CommentedMap()
        new_slide_player = CommentedMap()

        for event, elements in self.fc['slide_player'].items():
            self._migrate_slide_player_instance(elements, event, new_slide_player)

        self.fc['slide_player'] = new_slide_player

    def _migrate_slide_player_instance(self, elements, event, new_slide_player):
        self.log.debug("Converting '%s' display_elements to widgets",
                       event)

        if isinstance(elements, dict):
            elements = [elements]
        display, expire, persist, priority, slide, transition = self._get_slide_player_attributes(elements)
        elements = self._migrate_elements(elements, display)
        if not slide:
            slide = V4Migrator._get_slide_name(display)
        new_slide_player[event] = CommentedMap()
        new_slide_player[event][slide] = CommentedMap()
        self.log.debug("Adding slide:%s to slide_player:%s", slide,
                       event)
        if transition:
            self.log.debug("Moving transition: from slide: to "
                           "slide_player:")
            new_slide_player[event][slide]['transition'] = transition[0]
            new_slide_player[event][slide].ca.items['transition'] = (
                transition[1])
        if display:
            self.log.debug("Setting slide_player:target: to '%s'",
                           display)
            new_slide_player[event][slide]['target'] = display
        if expire:
            self.log.debug("Setting slide_player:expire: to '%s'",
                           expire)
            new_slide_player[event][slide]['expire'] = expire
        if priority:
            self.log.debug("Setting slide_player:priority: to '%s'",
                           priority)
            new_slide_player[event][slide]['priority'] = priority
        if persist:
            self.log.debug("Setting slide_player:persist: to '%s'",
                           persist)
            new_slide_player[event][slide]['persist'] = persist
        if not new_slide_player[event][slide]:
            new_slide_player[event] = slide
        self.log.debug("Creating slide: '%s' with %s migrated "
                       "widget(s)", slide, len(elements))
        self.fc['slides'][slide] = elements

    def _get_slide_player_attributes(self, elements):
        display = None
        transition = None
        expire = None
        priority = None
        persist = None
        slide = None

        for element in elements:
            if 'display' in element:
                self.log.debug("Converting display: to target:")
                display = element['display']
                del element['display']

            if 'transition' in element:
                transition = (element['transition'],
                              element.ca.items.get('transition', None))
                del element['transition']

            if 'expire' in element:
                expire = element['expire']
                del element['expire']

            if 'slide_priority' in element:
                priority = element['slide_priority']
                del element['slide_priority']

            if 'priority' in element:
                priority = element['priority']
                del element['priority']

            if 'persist_slide' in element:
                persist = element['persist_slide']
                del element['persist_slide']

            if 'slide_name' in element:
                slide = element['slide_name']
                del element['slide_name']

            if 'clear_slide' in element:
                del element['clear_slide']
        return display, expire, persist, priority, slide, transition

    def _migrate_shots(self):
        if 'shots' not in self.fc:
            return

        for v in self.fc['shots'].values():
            self._convert_show_call_to_tokens(v)

    def _create_window_slide(self):
        if 'window' in self.fc and 'elements' in self.fc['window']:
            elements = self.fc['window']['elements']

            if isinstance(elements, dict):
                elements = [elements]

            if 'slides' not in self.fc:
                self.log.debug("Creating 'slides:' section")
                self.fc['slides'] = CommentedMap()

            slide_name = V4Migrator._get_slide_name('window')

            self.log.debug("Creating slide: %s with %s display widget(s) from "
                           "the old window: config", slide_name, len(elements))

            self.log.debug("Adding '%s' slide", slide_name)
            self.fc['slides'][slide_name] = CommentedMap()
            self.fc['slides'][slide_name] = (
                self._migrate_elements(elements, 'window'))

            YamlRoundtrip.del_key_with_comments(self.fc['window'], 'elements',
                                                self.log)

            if 'slide_player' not in self.fc:
                self.fc['slide_player'] = CommentedMap()
                self.log.debug("Creating slide_player: section")

            self.log.debug("Creating slide_player:machine_reset_phase3: entry"
                           "to show slide '%s' on boot", slide_name)
            self.fc['slide_player']['machine_reset_phase_3'] = CommentedMap()
            self.fc['slide_player']['machine_reset_phase_3'][slide_name] = \
                CommentedMap()
            self.fc['slide_player']['machine_reset_phase_3'][slide_name][
                'target'] = 'window'

    def _migrate_sound_system(self):
        # convert stream track to regular track
        try:
            stream_track_name = self.fc['sound_system']['stream']['name']
            self.fc['sound_system']['tracks'][stream_track_name] = (
                CommentedMap())
            self.fc['sound_system']['tracks'][stream_track_name]['volume'] = 0.5
            self.fc['sound_system']['tracks'][stream_track_name]['simultaneous_sounds'] = 1
            self.log.debug('Converting stream: audio track to normal track')
        except KeyError:
            pass

        try:
            old_buffer = self.fc['sound_system']['buffer']
        except KeyError:
            old_buffer = None

        try:
            self.fc['sound_system']['buffer'] = 2048
            if old_buffer:
                self.fc['sound_system'].yaml_add_eol_comment(
                    'previous value was {}'.format(old_buffer), 'buffer')
                self.log.debug("Setting sound_system:buffer: to '2048'. "
                               "(Was %s)", old_buffer)
        except KeyError:
            pass

    def _migrate_fonts(self):
        # Fonts to widget_styles was already renamed, now update contents
        if 'widget_styles' in self.fc:
            self.log.debug("Converting widget_styles: from the old fonts: "
                           "settings")
            for settings in self.fc['widget_styles'].values():
                YamlRoundtrip.rename_key('size', 'font_size', settings,
                                         self.log)
                YamlRoundtrip.rename_key('file', 'font_name', settings,
                                         self.log)
                YamlRoundtrip.rename_key('crop_top', 'adjust_top', settings,
                                         self.log)
                YamlRoundtrip.rename_key('crop_bottom', 'adjust_bottom',
                                         settings, self.log)

                if 'font_name' in settings:
                    self.log.debug("Converting font_name: from file to name")
                    settings['font_name'] = os.path.splitext(
                        settings['font_name'])[0]

        if self.base_name == V4Migrator.MAIN_CONFIG_FILE:
            if 'widget_styles' not in self.fc:
                self.log.debug("Creating old default font settings as "
                               "widget_styles: section")
                self.fc['widget_styles'] = self._get_old_default_widget_styles()

            else:
                for k, v in self._get_old_default_widget_styles().items():
                    if k not in self.fc['widget_styles']:
                        self.fc['widget_styles'][k] = v

                    self.log.debug("Merging old built-in font settings '%s' "
                                   "into widget_styles: section", k)

    def _migrate_asset_defaults(self):
        # convert asset_defaults to assets:
        if 'asset_defaults' in self.fc:
            self.log.debug('Renaming key: asset_defaults -> assets:')
            YamlRoundtrip.rename_key('asset_defaults', 'assets', self.fc,
                                     self.log)

            assets = self.fc['assets']

            if 'animations' in assets:
                self.log.debug("Converting assets:animations to assets:images")
                if 'images' in assets:
                    self.log.debug("Merging animations: into current "
                                   "asset:images:")
                    YamlRoundtrip.copy_with_comments(assets, 'animations',
                                                     assets, 'images',
                                                     True, self.log)
                else:
                    YamlRoundtrip.rename_key('animations', 'images', assets,
                                             self.log)
                YamlRoundtrip.del_key_with_comments(self.fc, 'animations',
                                                    self.log)

            if 'movies' in assets:
                YamlRoundtrip.rename_key('movies', 'videos', assets, self.log)

            if 'images' in assets:
                self.log.debug("Converting assets:images:")

                for settings in assets['images'].values():
                    YamlRoundtrip.del_key_with_comments(settings, 'target',
                                                        self.log)

            if 'sounds' in assets:
                self.log.debug("Converting assets:sounds:")

                # for asset, settings in assets['sounds'].items():
                #     pass  # todo

    def _migrate_animation_assets(self):
        if 'animations' in self.fc:
            self.log.debug("Converting assets:animations to assets:images")
            if 'images' in self.fc:
                self.log.debug("Merging animations: into current "
                               "asset:images:")

                YamlRoundtrip.copy_with_comments(self.fc, 'animations',
                                                 self.fc, 'images',
                                                 True, self.log)

            else:
                YamlRoundtrip.rename_key('animations', 'images', self.fc,
                                         self.log)

    def _migrate_show_player(self):
        if 'show_player' not in self.fc:
            return

        temp_show_player = deepcopy(self.fc['show_player'])

        self.log.debug("Migrating show_player: section")

        for event, actions in self.fc['show_player'].items():
            if not isinstance(actions, list):
                actions = [actions]
            this_events_shows = CommentedMap()

            for action in actions:
                if 'show' in action:
                    show_name = action.pop('show')
                    this_events_shows[show_name] = action
                if 'action' in action and action['action'] == 'start':
                    del action['action']

            temp_show_player[event] = this_events_shows

        del self.fc['show_player']  # do not want to delete comments
        for event, shows in temp_show_player.items():
            self._add_to_show_player(event, shows)

    def _migrate_light_player(self):
        # light_player: section in v3 was used for both playing light scripts
        # and playing shows

        if 'light_player' not in self.fc:
            return

        self.log.debug("Migrating light_player: section")

        for event, actions in self.fc['light_player'].items():
            if not isinstance(actions, list):
                actions = [actions]
            this_events_shows = CommentedMap()

            for dummy_i, action in enumerate(actions):

                if 'show' in action:
                    show_name = action.pop('show')

                elif 'script' in action:
                    show_name = action.pop('script')

                elif 'key' in action:
                    show_name = action.pop('key')

                else:
                    continue

                if 'action' in action and action['action'] == 'start':
                    del action['action']

                this_events_shows[show_name] = action

            self._add_to_show_player(event, this_events_shows)

        YamlRoundtrip.del_key_with_comments(self.fc, 'light_player', self.log)

    def _migrate_shot_profiles(self):
        if 'shot_profiles' not in self.fc:
            return

        for settings in self.fc['shot_profiles'].values():
            if 'states' in settings:
                for dummy_i, state_settings in enumerate(settings['states']):
                    if 'loops' in state_settings and state_settings['loops']:
                        state_settings['loops'] = -1
                    YamlRoundtrip.rename_key('light_script', 'show',
                                             state_settings)

    def _add_to_show_player(self, event, show_dict):
        for settings in show_dict.values():
            if 'loops' in settings:
                if settings['loops']:
                    settings['loops'] = -1
                else:
                    settings['loops'] = 0

            self._convert_show_call_to_tokens(settings)

        if 'show_player' not in self.fc:
            self.log.debug("Creating show_player: section")
            self.fc['show_player'] = CommentedMap()

        if event not in self.fc['show_player']:
            self.log.debug("Updating show_player: content")
            self.fc['show_player'][event] = CommentedMap()
            self.fc['show_player'].ca.items[event] = show_dict.ca.items

        try:
            self.fc['show_player'][event].update(show_dict)
            self.fc['show_player'].ca.items[event] = (
                self.fc['light_player'].ca.items[event])
        except KeyError:
            pass

    @classmethod
    def _convert_show_call_to_tokens(cls, settings):
        token_list = ['light', 'lights', 'leds', 'led']

        for token in token_list:
            if token in settings:
                if 'show_tokens' not in settings:
                    settings['show_tokens'] = CommentedMap()

                YamlRoundtrip.copy_with_comments(settings, token,
                                                 settings['show_tokens'],
                                                 token, True)

    def _migrate_assets(self, section_name):
        if section_name in self.fc:

            keys_to_keep = set(self.mpf_config_spec[section_name].keys())
            empty_entries = set()

            self.log.debug("Converting %s: section", section_name)

            if self.fc[section_name]:

                for name, settings in self.fc[section_name].items():
                    self.log.debug("Converting %s:%s:", section_name, name)
                    if isinstance(settings, dict):
                        keys = set(settings.keys())
                        keys_to_remove = keys - keys_to_keep

                        for key in keys_to_remove:
                            YamlRoundtrip.del_key_with_comments(settings, key,
                                                                self.log)

                    if not settings:
                        self.log.debug("%s:%s: is now empty. Will remove it.",
                                       section_name, name)
                        empty_entries.add(name)

                for name in empty_entries:
                    YamlRoundtrip.del_key_with_comments(self.fc[section_name],
                                                        name, self.log)

            if not self.fc[section_name]:
                self.log.debug("%s: is now empty. Will remove it.",
                               section_name)
                YamlRoundtrip.del_key_with_comments(self.fc, section_name,
                                                    self.log)

    def _migrate_elements(self, elements, display=None):
        # takes a list of elements, returns a list of widgets
        if isinstance(elements, dict):
            elements = [elements]

        non_widgets = list()

        for i, element in enumerate(elements):
            elements[i] = self._element_to_widget(element, display)
            if not elements[i]:
                non_widgets.append(elements[i])

        for nw in non_widgets:
            elements.remove(nw)
            # todo do something with these?

        return elements

    @classmethod
    def _get_width_and_height_for_display(cls, element, display):
        # Figure out which display we're working with so we can get the
        # size to update the positions later. This could be target or
        # display, since this meth is called from a few different places
        if 'target' in element:
            display = element['target']
        elif 'display' in element:
            display = element['display']

        if not display:
            display = V4Migrator.default_display

        if display:
            width, height = V4Migrator.displays[display]
        elif V4Migrator.WIDTH and V4Migrator.HEIGHT:
            width = V4Migrator.WIDTH
            height = V4Migrator.HEIGHT
        else:
            raise ValueError("Unable to auto-detect display with and height. "
                             "Run the migrator again with the -h and -w "
                             "options to manually specific width and height")

        return width, height

    def _migrate_layer(self, element):
        # Migrate layer
        YamlRoundtrip.rename_key('layer', 'z', element, self.log)
        YamlRoundtrip.rename_key('h_pos', 'anchor_x', element, self.log)
        YamlRoundtrip.rename_key('v_pos', 'anchor_y', element, self.log)
        YamlRoundtrip.rename_key('font', 'style', element, self.log)
        YamlRoundtrip.rename_key('shade', 'brightness', element, self.log)
        YamlRoundtrip.del_key_with_comments(element, 'pixel_spacing', self.log)
        YamlRoundtrip.del_key_with_comments(element, 'antialias', self.log)
        YamlRoundtrip.del_key_with_comments(element, 'thickness', self.log)
        YamlRoundtrip.del_key_with_comments(element, 'bg_shade', self.log)
        YamlRoundtrip.del_key_with_comments(element, 'slide', self.log)
        return element

    @classmethod
    def _format_anchor_and_value(cls, anchor, value):
        if value < 0:
            return '{}{}'.format(anchor, value)

        return '{}+{}'.format(anchor, value)

    def _migrate_element_y_and_anchor(self, element, display, height):
        if 'y' in element:
            old_y = element['y']
            element['y'] *= -1
        else:
            old_y = 'None'

        if 'anchor_y' not in element and 'y' in element:
            element['anchor_y'] = 'top'

        try:
            if 'anchor_y' not in element or element['anchor_y'] == 'bottom':
                element['y'] = self._format_anchor_and_value('bottom', element['y'])
            elif element['anchor_y'] in ('middle', 'center'):
                element['y'] = self._format_anchor_and_value('middle', element['y'])
            elif element['anchor_y'] == 'top':
                element['y'] = self._format_anchor_and_value('top', element['y'])

            self.log.debug("Changing y:%s to y:%s (Based on anchor_y:%s"
                           "and %s height:%s)", old_y, element['y'],
                           element['anchor_y'], display, height)

        except KeyError:
            pass

        try:
            if element['anchor_y'] in ('middle', 'center'):
                YamlRoundtrip.del_key_with_comments(element, 'anchor_y',
                                                    self.log)

        except KeyError:
            pass

        if ('anchor_y' in element and 'y' not in element and
                element['anchor_y'] != 'middle'):
            element['y'] = element['anchor_y']
        return element

    def _migrate_element_x_and_anchor(self, element, display, width):
        if 'x' in element:
            old_x = element['x']
        else:
            old_x = 'None'

        if 'anchor_x' not in element and 'x' in element:
            element['anchor_x'] = 'left'

        try:
            if 'anchor_x' not in element or element['anchor_x'] == 'left':
                element['x'] = self._format_anchor_and_value('left', element['x'])

            elif element['anchor_x'] in ('middle', 'center'):
                element['x'] = self._format_anchor_and_value('center', element['x'])

            elif element['anchor_x'] == 'right':
                element['x'] = self._format_anchor_and_value('right', element['x'])

            self.log.debug("Changing x:%s to x:%s (Based on anchor_x:%s"
                           "and %s width:%s)", old_x, element['x'],
                           element['anchor_x'], display, width)

        except KeyError:
            pass

        try:
            if element['anchor_x'] in ('middle', 'center'):
                YamlRoundtrip.del_key_with_comments(element, 'anchor_x',
                                                    self.log)
        except KeyError:
            pass

        if ('anchor_x' in element and 'x' not in element and
                element['anchor_x'] != 'center'):
            element['x'] = element['anchor_x']
        return element

    def _element_to_widget(self, element, display):
        # takes an element dict, returns a widget dict

        width, height = self._get_width_and_height_for_display(element, display)

        try:
            element_type = element['type'].lower()

        except KeyError:
            return False

        element = self._migrate_layer(element)

        type_map = dict(virtualdmd='dmd',
                        text='text',
                        shape='shape',
                        animation='animation',
                        image='image',
                        movie='video',
                        character_picker='character_picker',
                        entered_chars='entered_chars')

        # Migrate the element type
        element['type'] = type_map[element_type]

        self.log.debug('Converting "%s" display_element to "%s" widget',
                       element_type, element['type'])

        if element_type == 'text':
            YamlRoundtrip.rename_key('size', 'font_size', element, self.log)

        if element_type != 'dmd':
            YamlRoundtrip.del_key_with_comments(element, 'bg_color', self.log)

        if element_type == 'virtualdmd' and V4Migrator.color_dmd:
            YamlRoundtrip.del_key_with_comments(element, 'pixel_color',
                                                self.log)
            self.log.debug('Changing widget type from "dmd" to "color_dmd"')
            element['type'] = 'color_dmd'

        element = self._migrate_element_x_and_anchor(element, display, height)
        element = self._migrate_element_y_and_anchor(element, display, width)

        if element_type == 'animation':
            element = self._migrate_animation(element)
        elif element_type == 'shape':
            element = self._migrate_shape(element)

        if 'decorators' in element:
            element = self._migrate_decorators(element, 'decorators',
                                               'animations')

        if 'cursor_decorators' in element:
            element = self._migrate_decorators(element, 'cursor_decorators',
                                               'cursor_animations')

        if 'color' in element:
            element['color'] = self._get_color(element['color'])

        if 'movie' in element:
            YamlRoundtrip.rename_key('movie', 'video', element, self.log)

            if 'repeat' in element:  # indented on purpose
                YamlRoundtrip.rename_key('repeat', 'loop', element, self.log)
            if 'loops' in element:  # indented on purpose
                YamlRoundtrip.rename_key('loops', 'loop', element, self.log)

        self._convert_tokens(element)

        return element

    def _get_color(self, color):
        color_tuple = RGBColor.hex_to_rgb(color)

        for color_name, val in NAMED_RGB_COLORS.items():
            if color_tuple == val:
                self.log.debug("Converting hex color '%s' to named color "
                               "'%s'", color, color_name)
                return color_name

        return color

    def _migrate_shape(self, element):
        if element['shape'] == 'box':
            self.log.debug("Converting 'box' display_element to 'rectangle' "
                           "widget")
            element['type'] = 'rectangle'
            del element['shape']

        elif element['shape'] == 'line':
            self.log.debug("Converting 'line' display_element to 'line' widget")
            element['type'] = 'line'
            del element['shape']

            element['points'] = (element.get('x', 0),
                                 element.get('y', 0),
                                 element.get('x', 0) + element['height'],
                                 element.get('y', 0) + element['width'])
        return element

    def _migrate_animation(self, element):
        self.log.debug("Converting 'animation' display_element to animated "
                       "'image' widget")
        element['type'] = 'image'
        YamlRoundtrip.rename_key('play_now', 'auto_play', element, self.log)
        YamlRoundtrip.rename_key('animation', 'image', element, self.log)

        element.pop('drop_frames', None)

        self.log.debug('Converting animated image loops: setting')
        if element['loops']:
            element['loops'] = -1
        else:
            element['loops'] = 0

        return element

    def _migrate_decorators(self, element, old_key, new_key):
        self.log.debug("Converting display_element blink decorator to widget "
                       "animation")
        decorator = element[old_key]

        element[new_key] = CommentedMap()
        element[new_key]['show_slide'] = CommentedSeq()

        on_dict = CommentedMap()
        on_dict['property'] = 'opacity'
        on_dict['value'] = 1
        on_dict['duration'] = str(decorator.get('on_secs', .5)) + 's'

        element[new_key]['show_slide'].append(on_dict)

        off_dict = CommentedMap()
        off_dict['property'] = 'opacity'
        off_dict['value'] = 0
        off_dict['duration'] = str(decorator.get('off_secs', .5)) + 's'
        off_dict['repeat'] = True

        element[new_key]['show_slide'].append(off_dict)

        del element[old_key]

        return element

    def _migrate_light_scripts(self):
        if 'light_scripts' not in self.fc:
            return

        YamlRoundtrip.rename_key('light_scripts', 'shows', self.fc, self.log)

        for show_contents in self.fc['shows'].values():
            self._convert_tocks_to_time(show_contents)

            for step in show_contents:

                if 'color' in step:
                    step['color'] = self._get_color(step['color'])
                    if len(str(step['color'])) > 2:
                        YamlRoundtrip.rename_key('color', '(leds)', step,
                                                 self.log)
                        step['leds'] = CommentedMap()
                        YamlRoundtrip.copy_with_comments(step, '(leds)',
                                                         step['leds'], '(leds)', True, self.log)
                    else:
                        YamlRoundtrip.rename_key('color', '(lights)', step,
                                                 self.log)
                        step['lights'] = CommentedMap()
                        YamlRoundtrip.copy_with_comments(step, '(lights)',
                                                         step['lights'], '(lights)', True, self.log)

    def _migrate_switches(self):
        if 'switches' not in self.fc:
            return

        for switch_settings in self.fc['switches'].values():
            YamlRoundtrip.rename_key('activation_events',
                                     'events_when_activated',
                                     switch_settings, self.log)
            YamlRoundtrip.rename_key('deactivation_events',
                                     'events_when_deactivated',
                                     switch_settings, self.log)

            if 'debounce' in switch_settings:
                if switch_settings['debounce']:
                    switch_settings['debounce'] = 'normal'
                else:
                    switch_settings['debounce'] = 'quick'

    @classmethod
    def _get_old_default_widget_styles(cls):
        # these are from MPF 0.21, but they are in the new v4 format
        widget_styles = '''
          default:
            font_name: Quadrit
            font_size: 10
            adjust_top: 2
            adjust_bottom: 3
          space title huge:
            font_name: DEADJIM
            font_size: 29
            antialias: true
            adjust_top: 3
            adjust_bottom: 3
          space title:
            font_name: DEADJIM
            font_size: 21
            antialias: true
            adjust_top: 2
            adjust_bottom: 3
          medium:
            font_name: pixelmix
            font_size: 8
            adjust_top: 1
            adjust_bottom: 1
          small:
            font_name: smallest_pixel-7
            font_size: 9
            adjust_top: 2
            adjust_bottom: 3
          tall title:
            font_name: big_noodle_titling
            font_size: 20
        '''

        return yaml.load(widget_styles, Loader=MpfRoundTripLoader)

    def _migrate_sound_player(self):
        if 'sound_player' not in self.fc:
            return

        self.log.debug("Migrating sound_player: section")

        temp_sound_player = CommentedMap()

        for settings in self.fc['sound_player'].values():
            this_sound = settings.pop('sound')
            play_events = settings.pop('start_events', None)
            stop_events = settings.pop('stop_events', None)

            play_events = Util.string_to_event_list(play_events)
            stop_events = Util.string_to_event_list(stop_events)

            for event in play_events:
                self._add_to_sound_player(temp_sound_player, event, this_sound,
                                          deepcopy(settings))

            for event in stop_events:
                self._add_to_sound_player(temp_sound_player, event, this_sound,
                                          dict(action='stop'))

        self.fc['sound_player'] = temp_sound_player

    @classmethod
    def _add_to_sound_player(cls, sound_player, event, sound, settings):
        if event not in sound_player:
            if settings:
                sound_player[event] = CommentedMap()
            elif not settings:
                sound_player[event] = sound
                return
        elif isinstance(sound_player[event], str):
            old_sound = sound_player[event]
            sound_player[event] = CommentedMap()
            sound_player[event][old_sound] = CommentedMap()
            sound_player[event][old_sound]['action'] = 'play'

        # if we're here, we now have a commented map

        if settings:
            sound_player[event][sound] = settings
        else:
            sound_player[event][sound] = CommentedMap()
            sound_player[event][sound]['action'] = 'play'

    @staticmethod
    def _migrate_logic_block(settings):
        try:
            if 'reset_each_ball' in settings:
                reset_each_ball = bool(settings['reset_each_ball'])

                del settings['reset_each_ball']

                if reset_each_ball:
                    if 'reset_events' in settings:
                        if isinstance(settings['reset_events'], dict):
                            settings['reset_events']['ball_starting'] = 0
                        elif isinstance(settings['reset_events'], list):
                            settings['reset_events'].append('ball_starting')
                        elif isinstance(settings['reset_events'], str):
                            settings['reset_events'] += ', ball_starting'

                    else:
                        settings['reset_events'] = 'ball_starting'
        except TypeError:
            pass
        return settings

    def _migrate_logic_blocks(self):
        if 'logic_blocks' not in self.fc:
            return

        for lb_type in self.fc['logic_blocks'].keys():
            for settings in self.fc['logic_blocks'][lb_type].values():
                self._migrate_logic_block(settings)

    def is_show_file(self):
        """Verify we have a show file and that it's an old version."""
        return 'tocks' in self.fc[0]

    def _migrate_show_file(self):
        self.log.debug("Migrating show file: %s", self.file_name)

        show_name_stub = os.path.splitext(os.path.split(self.file_name)[1])[0]

        self._add_show_version()

        # Convert tocks to time
        self._convert_tocks_to_time(self.fc)

        # migrate the components in each step
        self.log.debug("Converting settings for each show step")

        slide_num = 0

        for i, step in enumerate(self.fc):

            self._remove_tags(step)

            if 'display' in step:
                self.log.debug("Show step %s: Converting 'display' section",
                               i + 1)

                found_transition = False
                for widget in step['display']:
                    if 'transition' in widget:
                        found_transition = True
                        break

                if found_transition:
                    step['display'] = CommentedMap(
                        widgets=self._migrate_elements(step['display']))

                    for widget in step['display']['widgets']:

                        self._convert_tokens(widget)

                        if 'transition' in widget:
                            YamlRoundtrip.copy_with_comments(
                                widget, 'transition', step['display'],
                                'transition', True, self.log)

                else:
                    step['display'] = self._migrate_elements(step['display'])
                    self._convert_tokens(step['display'])

                YamlRoundtrip.rename_key('display', 'slides', step)

                slide_num += 1
                old_slides = step['slides']
                step['slides'] = CommentedMap()
                step['slides']['{}_slide_{}'.format(show_name_stub,
                                                    slide_num)] = old_slides

        return True

    def _add_show_version(self):
        # Do a str.replace to preserve any spaces or comments in the header

        try:
            self.fc.ca.comment[1][0].value = ('#show_version=4\n{}'.format(
                self.fc.ca.comment[1][0].value))
        except TypeError:
            self.fc.yaml_set_start_comment('show_version=4\n')

        self.log.debug('Adding #show_version=4')

    def _convert_tocks_to_time(self, show_steps):
        self.log.debug('Converting "tocks:" to "time:" and cascading entries '
                       'to the next step (since time: is for the current '
                       'step versus tocks: being for the previous step)')
        previous_tocks = 0
        for i, step in enumerate(show_steps):
            previous_tocks = step['tocks']

            if not i:
                step['tocks'] = 0
            else:
                step['tocks'] = '+{}'.format(previous_tocks)

            YamlRoundtrip.rename_key('tocks', 'time', step, self.log)

        if len(show_steps) > 1:
            show_steps.append(CommentedMap())
            show_steps[-1]['time'] = '+{}'.format(previous_tocks)

        return show_steps

    def _remove_tags(self, dic):
        found = False
        for v in dic.values():
            if isinstance(v, dict):
                for k1 in v.keys():
                    if k1.startswith('tag|'):
                        YamlRoundtrip.rename_key(k1, k1.strip('tag|'), v)
                        found = True
                        break

        if found:
            self._remove_tags(dic)

    @classmethod
    def _convert_tokens(cls, dic):
        # converts % tokens to ()
        token_finder = re.compile("(?<=%)[a-zA-Z_0-9|]+(?=%)")

        if isinstance(dic, list):
            for step in dic:
                if 'text' in step:
                    for token in token_finder.findall(str(step['text'])):
                        step['text'] = step['text'].replace(
                            '%{}%'.format(token), '({})'.format(token))

        else:
            if 'text' in dic:
                for token in token_finder.findall(str(dic['text'])):
                    dic['text'] = dic['text'].replace(
                        '%{}%'.format(token), '({})'.format(token))


def migrate_file(file_name, file_content):
    """Migrate file."""
    return V4Migrator(file_name, file_content).migrate()



"""Trinamics StepRocker controller platform."""
import asyncio
import logging

from mpf.platforms.trinamics import TMCL

from mpf.platforms.trinamics.TMCL import TMCLDevice
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface

from mpf.core.platform import StepperPlatform


class TrinamicsStepRocker(StepperPlatform):

    """Supports the Trinamics Step Rocker via PySerial.

    Works with Trinamics Step Rocker.  TBD other 'TMCL' based steppers eval boards
    """

    def __init__(self, machine):
        """Initialise Trinamics Step Rocker platform."""
        super().__init__(machine)
        self.log = logging.getLogger("Trinamics StepRocker")
        self.log.debug("Configuring template hardware interface.")
        self.config = self.machine.config['trinamics_steprocker']
        self.platform = None
        self.features['tickless'] = True
        self.tmcl = None

    def __repr__(self):
        """Return string representation."""
        return '<Platform.TrinamicsStepRocker>'

    async def initialize(self):
        """Initialise trinamics steprocker platform."""
        await super().initialize()

        # validate our config (has to be in intialize since config_processor
        # is not read in __init__)
        self.config = self.machine.config_validator.validate_config("trinamics_steprocker", self.config)
        self.tmcl = TMCLDevice(self.config['port'], False)

    def stop(self):
        """Close serial."""
        if self.tmcl:
            self.tmcl.stop()
            self.tmcl = None

    async def configure_stepper(self, number: str, config: dict) -> "TrinamicsTMCLStepper":
        """Configure a smart stepper device in platform.

        Args:
        ----
            number: Number of the stepper.
            config (dict): Configuration of device
        """
        return TrinamicsTMCLStepper(number, config, self.tmcl, self.machine)

    @classmethod
    def get_stepper_config_section(cls):
        """Return config validator name."""
        return "steprocker_stepper_settings"


# pylint: disable-msg=too-many-instance-attributes
class TrinamicsTMCLStepper(StepperPlatformInterface):

    """A stepper on a TMCL based controller such as Trinamics StepRocker."""

    def __init__(self, number, config, tmcl_device, machine):
        """Initialise stepper."""
        self._pulse_div = 5     # tbd add to config
        self._ramp_div = 9      # tbd add to config
        self._clock_freq = 16000000.0
        self.config = config
        self.log = logging.getLogger('TMCL Stepper')
        self._mn = int(number)
        self.tmcl = tmcl_device
        self._move_current = int(2.55 * self.config['move_current'])    # percent to 0...255(100%)
        self._hold_current = int(2.55 * self.config['hold_current'])    # percent to 0...255(100%)
        self._microstep_per_fullstep = self.config['microstep_per_fullstep']
        self._fullstep_per_userunit = self.config['fullstep_per_userunit']
        self._velocity_limit = self._uu_to_velocity_cmd(self.config['velocity_limit'])
        self._acceleration_limit = self._uu_to_accel_cmd(self.config['acceleration_limit'])
        self.machine = machine
        self._homing_speed = self._uu_to_velocity_cmd(self.config['homing_speed'])

        self._set_important_parameters(self._velocity_limit, self._acceleration_limit,
                                       self._move_current, self._hold_current,
                                       self._get_micro_step_mode(self._microstep_per_fullstep), False)
        # apply pulse and ramp divisors as well
        self.tmcl.sap(self._mn, 154, self._pulse_div)
        self.tmcl.sap(self._mn, 153, self._ramp_div)
        self._homing_active = False

    # Public Stepper Platform Interface
    def home(self, direction):
        """Home an axis, resetting 0 position."""
        self.tmcl.rfs(self._mn, 'STOP')  # in case in progress
        self._set_home_parameters(direction)
        self.tmcl.rfs(self._mn, 'START')
        self._homing_active = True

    def set_home_position(self):
        """Tell the stepper that we are at the home position."""
        self.tmcl.sap(self._mn, 1, 0)

    def move_abs_pos(self, position):
        """Move axis to a certain absolute position."""
        microstep_pos = self._uu_to_microsteps(position)
        self.tmcl.mvp(self._mn, "ABS", microstep_pos)

    def move_rel_pos(self, position):
        """Move axis to a relative position."""
        microstep_rel = self._uu_to_microsteps(position)
        self.tmcl.mvp(self._mn, "REL", microstep_rel)

    def move_vel_mode(self, velocity):
        """Move at a specific velocity and direction (pos = clockwise, neg = counterclockwise)."""
        self._rotate(velocity)

    def current_position(self):
        """Return current position."""
        microsteps = self.tmcl.gap(self._mn, 1)
        return self._microsteps_to_uu(microsteps)

    def stop(self) -> None:
        """Stop stepper."""
        self.tmcl.mst(self._mn)

    async def wait_for_move_completed(self):
        """Wait until move completed."""
        while not self.is_move_complete():
            await asyncio.sleep(1 / self.config['poll_ms'])

    def is_move_complete(self) -> bool:
        """Return true if move is complete."""
        if self._homing_active:
            ret = self.tmcl.rfs(self._mn, 'STATUS')
            if ret != 0:  # This is reversed from manual but is how it works
                return False

            self._homing_active = False
            return True

        # check normal move status
        ret = self.tmcl.gap(self._mn, 8)
        if ret == 1:
            return True

        return False

    # Private Utility Functions
    @staticmethod
    def _get_micro_step_mode(microsteps_per_fullstep: int) -> int:
        return int({
            1: 0,
            2: 1,
            4: 2,
            8: 3,
            16: 4,
            32: 5,
            64: 6,
            128: 7,
            256: 8
        }.get(microsteps_per_fullstep, 0))

    def _uu_to_microsteps(self, userunits) -> float:
        return userunits * self._fullstep_per_userunit * self._microstep_per_fullstep

    def _microsteps_to_uu(self, microsteps) -> float:
        return microsteps / (self._fullstep_per_userunit * self._microstep_per_fullstep)

    def _to_velocity_cmd(self, microsteps_per_sec) -> int:
        ret = (microsteps_per_sec * 2 ** self._pulse_div * 2048.0 * 32.0) / self._clock_freq
        if abs(ret) > 2047:
            raise ValueError("Scaled Velocity too high, lower pulse_div")
        if ret < 1:
            raise ValueError("Scaled Velocity too low, raise pulse_div")
        return int(ret)

    def _to_acceleration_cmd(self, microsteps_per_ss) -> int:
        ret = (2 ** (self._pulse_div + self._ramp_div + 29) * microsteps_per_ss) / self._clock_freq ** 2
        if ret > 2047:
            raise ValueError("Acceleration too high, lower ramps_div")
        if ret < 1:
            raise ValueError("Acceleration too low, raise ramps_div")
        return int(ret)

    def _uu_to_velocity_cmd(self, user_unit):
        microsteps_per_sec = user_unit * self._fullstep_per_userunit * self._microstep_per_fullstep
        return self._to_velocity_cmd(microsteps_per_sec)

    def _uu_to_accel_cmd(self, user_unit):
        microsteps_per_sec = user_unit * self._fullstep_per_userunit * self._microstep_per_fullstep
        return self._to_acceleration_cmd(microsteps_per_sec)

    def _get_globals(self):
        ret = {}
        for key, value in TMCL.GLOBAL_PARAMETER.iteritems():
            bank, par, name, _, _ = key + value
            ret[name] = self.tmcl.ggp(bank, par)
        return ret

    def _get_parameters(self):
        retmotor = [{}, {}, {}]
        retsingle = {}
        for mn in range(3):
            for key, value in TMCL.AXIS_PARAMETER.iteritems():
                par, name, _, _ = (key,) + value
                if par not in TMCL.SINGLE_AXIS_PARAMETERS:
                    retmotor[mn][name] = self.tmcl.gap(mn, par)
                elif mn == 0:
                    retsingle[name] = self.tmcl.gap(mn, par)
        return retmotor, retsingle

    # pylint: disable-msg=too-many-arguments
    def _set_important_parameters(self, maxspeed=2000, maxaccel=2000,
                                  maxcurrent=72, standbycurrent=32,
                                  microstep_resolution=1, store=False):
        self.tmcl.sap(self._mn, 140, int(microstep_resolution))
        self.tmcl.sap(self._mn, 4, int(maxspeed))
        self.tmcl.sap(self._mn, 5, int(maxaccel))
        self.tmcl.sap(self._mn, 6, int(maxcurrent))
        self.tmcl.sap(self._mn, 7, int(standbycurrent))
        if not bool(store):
            return
        self.tmcl.stap(self._mn, 140)
        self.tmcl.stap(self._mn, 4)
        self.tmcl.stap(self._mn, 5)
        self.tmcl.stap(self._mn, 6)
        self.tmcl.stap(self._mn, 7)

    def _set_home_parameters(self, direction):
        # self.TMCL.sap(self._mn, 9,  ) #ref. switch status
        # self.TMCL.sap(self._mn, 10, ) #right limit switch status
        # self.TMCL.sap(self._mn, 11, ) #left limit switch status
        # self.TMCL.sap(self._mn, 12, ) #right limit switch disable
        # self.TMCL.sap(self._mn, 13, ) #left limit switch disable
        # self.TMCL.sap(self._mn, 141, ) #ref. switch tolerance
        # self.TMCL.sap(self._mn, 149, ) #soft stop flag
        self.tmcl.sap(self._mn, 194, self._homing_speed)    # referencing search speed
        if direction == 'clockwise':
            self.tmcl.sap(self._mn, 193, 8)     # ref. search mode
        elif direction == 'counterclockwise':
            self.tmcl.sap(self._mn, 193, 7)
        # self.TMCL.sap(self._mn, 195, ) #referencing switch speed
        # self.TMCL.sap(self._mn, 196, ) # distance end switches

    def _rotate(self, velocity):
        if velocity == 0:
            self.tmcl.mst(self._mn)     # motor stop
        if velocity > 0:
            self.tmcl.ror(self._mn, self._uu_to_velocity_cmd(velocity))
        else:
            self.tmcl.rol(self._mn, self._uu_to_velocity_cmd(abs(velocity)))
        return velocity



import unittest
from mpf.core.utility_functions import Util


class TestUtil(unittest.TestCase):

    def test_string_to_ms(self):
        self.assertEqual(86400000, Util.string_to_ms('1d'))
        self.assertEqual(3600000, Util.string_to_ms('1h'))
        self.assertEqual(60000, Util.string_to_ms('1m'))
        self.assertEqual(1000, Util.string_to_ms('1s'))
        self.assertEqual(1, Util.string_to_ms('1ms'))
        self.assertEqual(0, Util.string_to_ms(None))
        self.assertEqual(0, Util.string_to_ms(False))

    def test_keys_to_lower(self):
        inner_dict = dict(key1=1, Key2=2)
        outer_dict = dict(key3=1, Key4=2, Key5=inner_dict)

        # Test that keys become lower, including in nested dicts
        self.assertIn('Key2', inner_dict)
        self.assertIn('Key4', outer_dict)
        self.assertIn('Key5', outer_dict)
        transformed_dict = Util.keys_to_lower(outer_dict)
        self.assertIn('key4', transformed_dict)
        self.assertIn('key2', transformed_dict['key5'])

        # Test that empty non-dicts work
        transformed_dict = Util.keys_to_lower(None)
        self.assertIs(type(transformed_dict), dict)

        # Test that lists of dicts convert keys to lower
        my_list = Util.keys_to_lower([inner_dict, outer_dict])
        self.assertIn('key2', my_list[0])
        self.assertIn('key4', my_list[1])

    def test_string_to_list(self):
        my_string = '0, 1, 2, 3, 4, 5'
        result = Util.string_to_event_list(my_string)
        self.assertEqual(type(result), list)
        self.assertEqual(result[5], '5')

        my_string = ['0', '1', '2', '3', '4', '5']
        result = Util.string_to_event_list(my_string)
        self.assertEqual(type(result), list)
        self.assertEqual(result[5], '5')

        my_string = None
        result = Util.string_to_event_list(my_string)
        self.assertEqual(type(result), list)
        self.assertFalse(result)

        my_string = '0, 1, 2{not, split here}, 3, 4, 5'
        result = Util.string_to_event_list(my_string)
        self.assertEqual(type(result), list)
        self.assertEqual(result[1], '1')
        self.assertEqual(result[2], '2{not, split here}')
        self.assertEqual(result[3], '3')

    def test_list_of_lists(self):
        my_list = ['1, 2, 3, 4, 5', 'a, b, c, d, e']
        result = Util.list_of_lists(my_list)
        self.assertEqual(type(result), list)
        self.assertEqual(result[0][0], '1')
        self.assertEqual(result[1][4], 'e')

    def test_dict_merge(self):
        dict_a = dict(key1='val1', key2='val2', list1=[1, 2, 3])
        dict_b = dict(key3='val3', key4='val4', list1=[4, 5, 6])

        result = Util.dict_merge(dict_a, dict_b, combine_lists=True)
        self.assertEqual(result['key1'], 'val1')
        self.assertEqual(result['key2'], 'val2')
        self.assertEqual(result['key3'], 'val3')
        self.assertEqual(result['key4'], 'val4')
        self.assertEqual(result['list1'], [1, 2, 3, 4, 5, 6])

        result = Util.dict_merge(dict_a, dict_b, combine_lists=False)
        self.assertEqual(result['key1'], 'val1')
        self.assertEqual(result['key2'], 'val2')
        self.assertEqual(result['key3'], 'val3')
        self.assertEqual(result['key4'], 'val4')
        self.assertEqual(result['list1'], [4, 5, 6])

    def test_hex_to_string_list(self):
        result = Util.hex_string_to_list('00ff88')
        self.assertEqual(result, [0, 255, 136])

        result = Util.hex_string_to_list('00FF88')
        self.assertEqual(result, [0, 255, 136])

        result = Util.hex_string_to_list('00ff88aa')
        self.assertEqual(result, [0, 255, 136])

        result = Util.hex_string_to_list('00ff88', 4)
        self.assertEqual(result, [0, 0, 255, 136])

        self.assertRaises(ValueError, Util.hex_string_to_list, 'g0')

    def test_hex_string_to_int(self):
        result = Util.hex_string_to_int('ff')
        self.assertEqual(result, 255)

        result = Util.hex_string_to_int('ff', maxvalue=128)
        self.assertEqual(result, 128)

    def test_event_config_to_dict(self):
        config = dict()
        config['test_events_1'] = dict()
        config['test_events_1']['event1'] = '1s'
        config['test_events_1']['event2'] = '100ms'
        config['test_events_2'] = dict()
        config['test_events_2']['event3'] = '1s'
        config['test_events_2']['event4'] = '100ms'
        results = Util.event_config_to_dict(config)
        self.assertEqual(results['test_events_1']['event1'], '1s')

        # should config string to dict with vals of 0
        config = 'event1, event2'
        results = Util.event_config_to_dict(config)
        self.assertEqual(results['event1'], 0)

        # should config list to dict with vals of 0
        config = ['event1', 'event2']
        results = Util.event_config_to_dict(config)
        self.assertEqual(results['event1'], 0)

    def test_int_to_hex_string(self):
        result = Util.int_to_hex_string(255)
        self.assertEqual(result, 'FF')

        self.assertRaises(ValueError, Util.int_to_hex_string, 256)

    def test_pwm8_to_hex_string(self):
        self.assertEqual(Util.pwm8_to_hex_string(0), '00')
        self.assertEqual(Util.pwm8_to_hex_string(4), 'AA')
        self.assertRaises(ValueError, Util.pwm8_to_hex_string, 9)

    def test_pwm32_to_hex_string(self):
        self.assertEqual(Util.pwm32_to_hex_string(0), '00000000')
        self.assertEqual(Util.pwm32_to_hex_string(4), '02020202')
        self.assertRaises(ValueError, Util.pwm32_to_hex_string, 33)

    def test_pwm8_to_int(self):
        self.assertEqual(Util.pwm8_to_int(0), 0)
        self.assertEqual(Util.pwm8_to_int(4), 170)
        self.assertRaises(ValueError, Util.pwm8_to_int, 9)

    def test_pwm32_to_int(self):
        self.assertEqual(Util.pwm32_to_int(0), 0)
        self.assertEqual(Util.pwm32_to_int(4), 33686018)
        self.assertRaises(ValueError, Util.pwm32_to_int, 33)

    def test_power_to_on_off(self):
        self.assertEqual(Util.power_to_on_off(0), (0, 0))
        self.assertEqual(Util.power_to_on_off(0.1), (1, 9))
        self.assertEqual(Util.power_to_on_off(0.15), (3, 17))
        self.assertEqual(Util.power_to_on_off(0.5), (1, 1))
        self.assertEqual(Util.power_to_on_off(1), (1, 0))
        self.assertRaises(ValueError, Util.power_to_on_off, 9)

    def test_normalize_hex_string(self):
        result = Util.normalize_hex_string('ff00', 4)
        self.assertEqual(result, 'FF00')

        result = Util.normalize_hex_string('ff00', 6)
        self.assertEqual(result, '00FF00')

        result = Util.normalize_hex_string('c')
        self.assertEqual(result, '0C')

        result = Util.normalize_hex_string('4')
        self.assertEqual(result, '04')

    def test_bin_str_to_hex_str(self):
        result = Util.bin_str_to_hex_str('1111', 2)
        self.assertEqual(result, '0F')

    def test_is_hex_string(self):
        self.assertTrue(Util.is_hex_string('ff1234'))
        self.assertTrue(Util.is_hex_string('123456'))
        self.assertTrue(Util.is_hex_string('aabb00'))
        self.assertFalse(Util.is_hex_string('ffaagg'))
        self.assertFalse(Util.is_hex_string('hello'))
        self.assertFalse(Util.is_hex_string([1, 2, 3]))

    def test_get_from_dict(self):
        a = dict()
        a['b'] = dict()
        a['b']['c'] = dict()
        a['b']['c']['d'] = 1

        self.assertEqual(Util.get_from_dict(a, ['b', 'c', 'd']), 1)

    def test_set_in_dict(self):
        a = dict()
        a['b'] = dict()
        a['b']['c'] = dict()

        Util.set_in_dict(a, ['b', 'c'], 1)
        self.assertEqual(a['b']['c'], 1)

    def test_power_of_2(self):
        self.assertTrue(Util.is_power2(2))
        self.assertTrue(Util.is_power2(256))
        self.assertFalse(Util.is_power2(3))
        self.assertFalse(Util.is_power2(222))



"""Motor device."""
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor(_move_direction="move_direction", _target_position="target_position", _last_position="last_position")
class Motor(SystemWideDevice):

    """A motor which can be controlled using drivers."""

    __slots__ = ["_target_position", "_last_position", "type", "_move_direction"]

    config_section = 'motors'
    collection = 'motors'
    class_label = 'motor'

    def __init__(self, machine, name):
        """Initialise motor."""
        self._target_position = None
        self._last_position = None
        self.type = None
        self._move_direction = "stopped"
        super().__init__(machine, name)

    async def _initialize(self):
        await super()._initialize()
        self._target_position = self.config['reset_position']
        if self.config['reset_position'] not in self.config['position_switches']:
            self.raise_config_error("Reset position {} not in positions {}".format(
                self.config['reset_position'], self.config['position_switches']), 1)

        if not self.config['motor_left_output'] and not self.config['motor_right_output']:
            self.raise_config_error("Need either motor_left_output or motor_right_output", 2)

        if self.config['motor_left_output'] == self.config['motor_right_output']:
            self.raise_config_error("motor_left_output and motor_right_output need to be different", 3)

        if self.config['motor_left_output'] and self.config['motor_right_output']:
            self.type = "two_directions"
        else:
            self.type = "one_direction"

        for position, switch in self.config['position_switches'].items():
            self.machine.switch_controller.add_switch_handler_obj(switch, self._update_position,
                                                                  callback_kwargs={"position": position})

        # add handlers
        for event, position in self.config['go_to_position'].items():
            if position not in self.config['position_switches']:
                self.raise_config_error("Invalid position {} in go_to_position".format(position), 4)
            self.machine.events.add_handler(event, self.event_go_to_position, position=position)

        if self.config['include_in_ball_search']:
            self.machine.events.add_handler("ball_search_started",
                                            self._ball_search_start)
            self.machine.events.add_handler("ball_search_stopped",
                                            self._ball_search_stop)

    def _validate_last_position(self) -> bool:
        """Verify that at most one position switch is active."""
        active_position_switches = [(position, switch) for position, switch in
                                    self.config['position_switches'].items() if switch.state]
        if len(active_position_switches) > 1:
            self.warning_log("Found %s active position switches: %s. There should be only one position switch active "
                             "at a time.", len(active_position_switches),
                             active_position_switches)
            self.machine.service.add_technical_alert(
                self, "Multiple position switches are active: {}. Verify switches.".format(active_position_switches))
            return False

        return True

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self. reset()

    def reset(self):
        """Go to reset position."""
        self.go_to_position(self.config['reset_position'])

    @event_handler(10)
    def event_go_to_position(self, position=None, **kwargs):
        """Event handler for go_to_position event."""
        del kwargs
        if position is None:
            raise AssertionError("Got go_to_position event without position.")

        self.go_to_position(position)

    def go_to_position(self, position):
        """Move motor to a specific position."""
        self.log.info("Moving motor to position %s", position)
        self._target_position = position
        self._move_to_position(position)

    def _move_to_position(self, position):
        if not self._validate_last_position():
            self.warning_log("Will not move motor because multiple position switches are active.")
            self._stop_motor()
            return

        switch = self.config['position_switches'][position]
        # check if we are already in this position
        if self.machine.switch_controller.is_active(switch):
            # already in position
            self._reached_position(position)
            self._stop_motor()
        else:
            if self.type == "two_directions":
                if self._last_position:
                    assumed_position = self._last_position
                else:
                    active_position_switches = [position for position, switch in
                                                self.config['position_switches'].items()
                                                if switch.state]
                    if len(active_position_switches) == 1:
                        assumed_position = active_position_switches[0]
                        self.debug_log("Assuming position based on switches to be %s", assumed_position)
                    else:
                        assumed_position = None

                if assumed_position is None and \
                        list(self.config['position_switches']).index(self.config['reset_position']) == 0:
                    self._move_left()
                elif assumed_position is None:
                    self._move_right()
                elif list(self.config['position_switches']).index(assumed_position) > \
                        list(self.config['position_switches']).index(position):
                    self._move_left()
                else:
                    self._move_right()
            else:
                # not in position. start motor
                if self.config['motor_left_output']:
                    self._move_left()
                else:
                    self._move_right()

    def _update_position(self, position, **kwargs):
        """Handle that motor reached a certain position."""
        del kwargs
        first_known_position = self._last_position is None
        if not self._validate_last_position():
            self.warning_log("Will stop motor because multiple position switches are active.")
            self._stop_motor()
            self._last_position = None
            return

        self._last_position = position

        if position == self._target_position:
            self._reached_position(position)
        else:
            self.debug_log("Motor is at position %s", position)

            if self.type == "two_directions":
                # special case: initial position has been unknown and we reached our first position
                # we might have moved in the wrong direction so correct this now
                if first_known_position and self._move_direction == "right" and \
                        list(self.config['position_switches']).index(self._last_position) > \
                        list(self.config['position_switches']).index(self._target_position):
                    self._move_left()
                elif first_known_position and self._move_direction == "left" and \
                        list(self.config['position_switches']).index(self._last_position) < \
                        list(self.config['position_switches']).index(self._target_position):
                    self._move_right()
                elif list(self.config['position_switches']).index(position) in \
                        (0, len(self.config['position_switches']) - 1):
                    self.warning_log("Motor hit end switch %s unexpectedly. Stopping motor.", position)
                    self._stop_motor()

    def _reached_position(self, position):
        """Handle that motor handled its target position."""
        self.info_log("Motor reached position %s. Stopping motor.", position)
        self.machine.events.post("motor_{}_reached_{}".format(self.name, position))
        '''event: motor_(name)_reached_(position)

        desc: A motor device called (name) reached position (position)
        (device)
        '''

        # disable motor
        self._stop_motor()

    def _stop_motor(self):
        if self.config['motor_left_output']:
            self.config['motor_left_output'].disable()

        if self.config['motor_right_output']:
            self.config['motor_right_output'].disable()

        self._move_direction = "stopped"

    def _move_right(self):
        if self.config['motor_left_output']:
            self.config['motor_left_output'].disable()

        self.config['motor_right_output'].enable()
        self._move_direction = "right"

    def _move_left(self):
        if self.config['motor_right_output']:
            self.config['motor_right_output'].disable()

        self.config['motor_left_output'].enable()
        self._move_direction = "left"

    def _ball_search_start(self, **kwargs):
        del kwargs
        self._stop_motor()
        # simply enable motor. will move to old position afterwards.
        if self.config['motor_left_output']:
            self.config['motor_left_output'].enable()
        else:
            self.config['motor_right_output'].enable()

    def _ball_search_stop(self, **kwargs):
        del kwargs
        # move to last position
        self._move_to_position(self._target_position)



import sys

from mpf.tests.MpfTestCase import MpfTestCase, MagicMock, patch


class MockServer:

    async def stop(self, grace):
        pass

    async def wait_for_termination(self):
        pass


class TestVPE(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/vpe/'

    async def _connect_to_mock_client(self, service, port):
        self.service = service
        await self.simulator.connect(self.service)
        return MockServer()

    def _mock_loop(self):
        self.simulator.init_async()

    def setUp(self):
        if sys.version_info < (3, 6):
            self.skipTest("Test requires Python 3.6+")
            return

        try:
            from mpf.tests.vpe_simulator import VpeSimulation
            from mpf.platforms.visual_pinball_engine import platform_pb2
        except (SyntaxError, ImportError) as e:
            self.skipTest("Cannot import VPE simulator because {}".format(e))
            return

        self.simulator = VpeSimulation({"0": True, "3": False, "6": False})
        import mpf.platforms.visual_pinball_engine.visual_pinball_engine
        mpf.platforms.visual_pinball_engine.visual_pinball_engine.VisualPinballEnginePlatform.listen = self._connect_to_mock_client
        super().setUp()

    def get_platform(self):
        return False

    def test_vpe(self):
        from mpf.platforms.visual_pinball_engine import platform_pb2
        description = self.loop.run_until_complete(
            self.service.GetMachineDescription(platform_pb2.EmptyRequest(), None))
        self.assertEqual(len(description.switches), 3)
        self.assertEqual(len(description.coils), 3)
        self.assertEqual(len(description.lights), 2)

        self.assertTrue(platform_pb2.SwitchDescription(
            name="s_sling",
            hardware_number="0",
            switch_type="NO",
        ) in description.switches)

        self.assertTrue(platform_pb2.CoilDescription(
            name="c_flipper",
            hardware_number="1"
        ) in description.coils)

        self.assertTrue(platform_pb2.LightDescription(
            name="test_light1",
            hardware_channel_number="light-0",
            hardware_channel_color="WHITE"
        ) in description.lights)

        self.assertTrue(platform_pb2.DmdDescription(
            name="default",
            color_mapping=platform_pb2.DmdDescription.ColorMapping.BW,
            width=128,
            height=32
        ) in description.dmds)

        self.assertTrue(platform_pb2.DmdDescription(
            name="test_dmd",
            color_mapping=platform_pb2.DmdDescription.ColorMapping.RGB,
            width=128,
            height=32
        ) in description.dmds)

        self.assertSwitchState("s_sling", True)
        self.assertSwitchState("s_flipper", False)
        self.assertSwitchState("s_test", False)
        self.simulator.set_switch("6", True)
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test", True)

        self.simulator.set_switch("6", False)
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test", False)

        self.machine.lights["test_light1"].color("CCCCCC")
        self.advance_time_and_run(.1)
        self.assertAlmostEqual(0.8, self.simulator.lights["light-0"])

        self.machine.coils["c_flipper"].pulse()
        self.advance_time_and_run(.1)
        self.assertEqual("pulsed-10-1.0", self.simulator.coils["1"])

        self.machine.coils["c_flipper"].enable()
        self.advance_time_and_run(.1)
        self.assertEqual("enabled-10-1.0-1.0", self.simulator.coils["1"])

        self.machine.coils["c_flipper"].disable()
        self.advance_time_and_run(.1)
        self.assertEqual("disabled", self.simulator.coils["1"])

        self.machine.flippers["f_test"].enable()
        self.advance_time_and_run(.1)
        self.assertEqual(platform_pb2.ConfigureHardwareRuleRequest(
            coil_number="1", switch_number="3", pulse_ms=10, pulse_power=1.0, hold_power=1.0),
            self.simulator.rules["1-3"])

        self.machine.flippers["f_test"].disable()
        self.advance_time_and_run(.1)
        self.assertNotIn("1-3", self.simulator.rules)

        # test set frame to buffer
        frame = bytearray()
        for i in range(4096):
            frame.append(i % 256)

        self.machine.dmds["default"].update(bytes(frame))
        self.advance_time_and_run(.1)

        self.assertEqual((frame, 1.0), self.simulator.dmd_frames["default"])

        rgb_frame = bytearray()
        for i in range(128 * 32 * 3):
            rgb_frame.append(i % 256)

        self.machine.rgb_dmds["test_dmd"].update(bytes(rgb_frame))
        self.advance_time_and_run(.1)

        self.assertEqual((rgb_frame, 1.0), self.simulator.dmd_frames["test_dmd"])



"""Score queues for SS games."""
import asyncio
import math
from mpf.core.utility_functions import Util

from mpf.core.system_wide_device import SystemWideDevice


class ScoreQueue(SystemWideDevice):

    """Score queues for SS games.

    Add scores over time and play a lot of chimes.
    """

    config_section = 'score_queues'
    collection = 'score_queues'
    class_label = 'score_queue'

    __slots__ = ["_score_queue", "_score_queue_empty", "_score_task"]

    def __init__(self, machine, name):
        """Initialise ball lock."""
        super().__init__(machine, name)
        self._score_queue = asyncio.Queue()
        self._score_queue_empty = asyncio.Event()
        self._score_queue_empty.set()
        self._score_task = None

        self.machine.events.add_async_handler("player_turn_ending", self._block_player_end_if_scoring)

    async def _block_player_end_if_scoring(self, **kwargs):
        """Block player ending until scoring is done."""
        del kwargs
        await self._score_queue_empty.wait()

    async def _initialize(self):
        await super()._initialize()
        self._score_task = self.machine.clock.loop.create_task(self._handle_score_queue())
        self._score_task.add_done_callback(Util.raise_exceptions)

    def score(self, value, **kwargs):
        """Score a value via the queue."""
        del kwargs
        if not self.machine.game or not self.machine.game.player:
            self.warning_log("Trying to use score_queue without an active game or player")
            return
        self._score_queue_empty.clear()
        self._score_queue.put_nowait(value)

    def stop_device(self):
        """Stop queue."""
        if self._score_task:
            self._score_task.cancel()
            self._score_task = None

    async def _handle_score_queue(self):
        while True:
            score = await self._score_queue.get()
            self.debug_log("Scoring %s", score)
            while score > 0:
                # get the position of the highest digit
                digit_pos = int(math.floor(math.log10(score)))
                digit_score = int(math.pow(10, digit_pos))
                # score this amount
                self.machine.game.player[self.name] += digit_score
                # reduce the remaining amount
                score -= digit_score
                self.debug_log("Scoring %s on digit %s. Remaining: %s", digit_score, digit_pos, score)
                # wait if there is a chime for that digit
                if len(self.config['chimes']) >= digit_pos and self.config['chimes'][-(digit_pos + 1)]:
                    self.config['chimes'][-(digit_pos + 1)].pulse()
                    self.debug_log("Played chime for pos %s. Waiting %ss", digit_pos, self.config['delay'])
                    await asyncio.sleep(self.config['delay'])

            if self._score_queue.empty():
                self._score_queue_empty.set()



"""Interface for a light hardware devices."""
import abc
import asyncio
from asyncio import AbstractEventLoop
from functools import total_ordering

from typing import Any, Optional

from mpf.core.utility_functions import Util


@total_ordering     # type: ignore
class LightPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for a light in hardware platforms."""

    __slots__ = ["number"]

    def __init__(self, number: Any) -> None:
        """Initialise light."""
        self.number = number

    @abc.abstractmethod
    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Perform a fade to a brightness.

        Args:
        ----
            start_brightness: Brightness at start of fade.
            start_time: Timestamp when the fade started.
            target_brightness: Brightness at end of fade.
            target_time: Timestamp when the fade should finish.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_board_name(self):
        """Return the name of the board of this light."""
        raise NotImplementedError

    def is_successor_of(self, other):
        """Return true if this light is the direct successor of the other light passed as parameter."""
        raise NotImplementedError

    def get_successor_number(self):
        """Return the next light number after this light if possible.

        This is used for serial LED chains. Raise an error if this is not possible.
        """
        raise NotImplementedError

    def __lt__(self, other):
        """Order lights by their position on the hardware."""
        raise NotImplementedError

    def __repr__(self):
        """Return string representation."""
        return "<{} number={}>".format(self.__class__, self.number)


class LightPlatformDirectFade(LightPlatformInterface, metaclass=abc.ABCMeta):

    """Implement a light which can set fade and brightness directly."""

    __slots__ = ["loop", "task"]

    def __init__(self, number, loop: AbstractEventLoop) -> None:
        """Initialise light."""
        super().__init__(number)
        self.loop = loop
        self.task = None    # type: Optional[asyncio.Task]

    @abc.abstractmethod
    def get_max_fade_ms(self) -> int:
        """Return max fade time."""
        raise NotImplementedError()

    def get_fade_interval_ms(self) -> int:
        """Return max fade time."""
        return self.get_max_fade_ms()

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Perform a fade with either a asyncio task or with a single command."""
        max_fade_ms = self.get_max_fade_ms()
        current_time = self.loop.time()
        if target_time > 0:
            fade_ms = (target_time - current_time) / 1000.0
        else:
            fade_ms = -1

        if fade_ms > max_fade_ms:
            # we have to continue the fade later
            if self.task:
                self.task.cancel()
            self.task = self.loop.create_task(self._fade(start_brightness, start_time, target_brightness, target_time))
            self.task.add_done_callback(Util.raise_exceptions)
        else:
            self.set_brightness_and_fade(target_brightness, max(fade_ms, 0))

    async def _fade(self, start_brightness, start_time, target_brightness, target_time):
        max_fade_ms = self.get_max_fade_ms()
        interval = self.get_fade_interval_ms() / 1000
        while True:
            current_time = self.loop.time()
            target_fade_ms = int((target_time - current_time) * 1000.0)
            if target_fade_ms > max_fade_ms >= 0:
                fade_ms = max_fade_ms
                ratio = ((current_time + (max_fade_ms / 1000.0) - start_time) /
                         (target_time - start_time))
                brightness = start_brightness + (target_brightness - start_brightness) * ratio
            else:
                fade_ms = target_fade_ms
                brightness = target_brightness
            self.set_brightness_and_fade(brightness, max(fade_ms, 0))
            if target_fade_ms <= max_fade_ms:
                return
            await asyncio.sleep(interval)

    @abc.abstractmethod
    def set_brightness_and_fade(self, brightness: float, fade_ms: int) -> None:
        """Set the light to the specified brightness.

        Args:
        ----
            brightness: float of the brightness
            fade_ms: ms to fade the light

        Does not return anything.
        """
        raise NotImplementedError

    def stop(self):
        """Stop light."""
        if self.task:
            self.task.cancel()


class LightPlatformSoftwareFade(LightPlatformDirectFade, metaclass=abc.ABCMeta):

    """Implement a light which cannot fade on its own."""

    __slots__ = ["software_fade_ms"]

    def __init__(self, number, loop: AbstractEventLoop, software_fade_ms: int) -> None:
        """Initialise light with software fade."""
        super().__init__(number, loop)
        self.software_fade_ms = software_fade_ms

    def get_max_fade_ms(self) -> int:
        """Return max fade time."""
        return 0

    def get_fade_interval_ms(self) -> int:
        """Return software fade interval."""
        return self.software_fade_ms

    def set_brightness_and_fade(self, brightness: float, fade_ms: int):
        """Set brightness and ensure that fade is 0."""
        assert fade_ms == 0
        self.set_brightness(brightness)

    @abc.abstractmethod
    def set_brightness(self, brightness: float) -> None:
        """Set the light to the specified brightness.

        Args:
        ----
            brightness: float of the brightness

        Does not return anything.
        """
        raise NotImplementedError



"""Fast serial communicator."""
import asyncio
from distutils.version import StrictVersion

from mpf.core.utility_functions import Util

from mpf.platforms.base_serial_communicator import BaseSerialCommunicator

# Minimum firmware versions needed for this module
from mpf.platforms.fast.fast_io_board import FastIoBoard

DMD_MIN_FW = '0.88'
NET_MIN_FW = '0.88'
RGB_MIN_FW = '0.87'
IO_MIN_FW = '0.87'

# DMD_LATEST_FW = '0.88'
# NET_LATEST_FW = '0.90'
# RGB_LATEST_FW = '0.88'
# IO_LATEST_FW = '0.89'


class FastSerialCommunicator(BaseSerialCommunicator):

    """Handles the serial communication to the FAST platform."""

    ignored_messages = ['RX:P',  # RGB Pass
                        'SN:P',  # Network Switch pass
                        'SL:P',  # Local Switch pass
                        'LX:P',  # Lamp pass
                        'PX:P',  # Segment pass
                        'DN:P',  # Network driver pass
                        'DL:P',  # Local driver pass
                        'XX:F',  # Unrecognized command?
                        'R1:F',
                        'L1:P',
                        'GI:P',
                        'TL:P',
                        'TN:P',
                        'XO:P',  # Servo/Daughterboard Pass
                        'XX:U',
                        'XX:N'
                        ]

    __slots__ = ["dmd", "remote_processor", "remote_model", "remote_firmware", "max_messages_in_flight",
                 "messages_in_flight", "ignored_messages_in_flight", "send_ready", "write_task", "received_msg",
                 "send_queue"]

    def __init__(self, platform, port, baud):
        """Initialise communicator.

        Args:
        ----
            platform(mpf.platforms.fast.fast.HardwarePlatform): the fast hardware platform
            port: serial port
            baud: baud rate
        """
        self.dmd = False

        self.remote_processor = None
        self.remote_model = None
        self.remote_firmware = 0.0
        self.max_messages_in_flight = 10
        self.messages_in_flight = 0
        self.ignored_messages_in_flight = {b'-N', b'/N', b'/L', b'-L'}

        self.send_ready = asyncio.Event()
        self.send_ready.set()
        self.write_task = None

        self.received_msg = b''

        self.send_queue = asyncio.Queue()

        super().__init__(platform, port, baud)

    def stop(self):
        """Stop and shut down this serial connection."""
        self.write_task.cancel()
        super().stop()

    async def _read_with_timeout(self, timeout):
        msg_raw = await asyncio.wait([self.readuntil(b'\r')], timeout=timeout)
        if not msg_raw[0]:
            msg_raw[1].pop().cancel()
            return ""
        element = msg_raw[0].pop()
        return (await element).decode()

    async def _identify_connection(self):
        """Identify which processor this serial connection is talking to."""
        # keep looping and wait for an ID response

        msg = ''

        # send enough dummy commands to clear out any buffers on the FAST
        # board that might be waiting for more commands
        self.writer.write(((' ' * 256 * 4) + '\r').encode())

        while True:
            self.platform.debug_log("Sending 'ID:' command to port '%s'",
                                    self.port)
            self.writer.write('ID:\r'.encode())
            msg = await self._read_with_timeout(.5)

            # ignore XX replies here.
            while msg.startswith('XX:'):
                msg = await self._read_with_timeout(.5)

            if msg.startswith('ID:'):
                break

            await asyncio.sleep(.5)

        # examples of ID responses
        # ID:DMD FP-CPU-002-1 00.87
        # ID:NET FP-CPU-002-2 00.85
        # ID:RGB FP-CPU-002-2 00.85

        try:
            self.remote_processor, self.remote_model, self.remote_firmware = (
                msg[3:].split())
        except ValueError:
            self.remote_processor, self.remote_model, = msg[3:].split()

        self.platform.log.info("Connected! Processor: %s, "
                               "Board Type: %s, Firmware: %s",
                               self.remote_processor, self.remote_model,
                               self.remote_firmware)

        self.machine.variables.set_machine_var("fast_{}_firmware".format(self.remote_processor.lower()),
                                               self.remote_firmware)
        '''machine_var: fast_(x)_firmware

        desc: Holds the version number of the firmware for the processor on
        the FAST Pinball controller that's connected. The "x" is replaced with
        either "dmd", "net", or "rgb", one for each processor that's attached.
        '''

        self.machine.variables.set_machine_var("fast_{}_model".format(self.remote_processor.lower()), self.remote_model)

        '''machine_var: fast_(x)_model

        desc: Holds the model number of the board for the processor on
        the FAST Pinball controller that's connected. The "x" is replaced with
        either "dmd", "net", or "rgb", one for each processor that's attached.
        '''

        if self.remote_processor == 'DMD':
            min_version = DMD_MIN_FW
            # latest_version = DMD_LATEST_FW
            self.dmd = True
            self.max_messages_in_flight = self.platform.config['dmd_buffer']
            self.platform.debug_log("Setting DMD buffer size: %s",
                                    self.max_messages_in_flight)
        elif self.remote_processor == 'NET':
            min_version = NET_MIN_FW
            # latest_version = NET_LATEST_FW
            self.max_messages_in_flight = self.platform.config['net_buffer']
            self.platform.debug_log("Setting NET buffer size: %s",
                                    self.max_messages_in_flight)
        elif self.remote_processor == 'RGB':
            min_version = RGB_MIN_FW
            # latest_version = RGB_LATEST_FW
            self.max_messages_in_flight = self.platform.config['rgb_buffer']
            self.platform.debug_log("Setting RGB buffer size: %s",
                                    self.max_messages_in_flight)
        else:
            raise AttributeError("Unrecognized FAST processor type: {}".format(self.remote_processor))

        if StrictVersion(min_version) > StrictVersion(self.remote_firmware):
            raise AssertionError('Firmware version mismatch. MPF requires'
                                 ' the {} processor to be firmware {}, but yours is {}'.
                                 format(self.remote_processor, min_version, self.remote_firmware))

        if self.remote_processor == 'NET' and self.platform.machine_type == 'fast':
            await self.query_fast_io_boards()

        self.platform.register_processor_connection(self.remote_processor, self)

        self.write_task = self.machine.clock.loop.create_task(self._socket_writer())
        self.write_task.add_done_callback(Util.raise_exceptions)

    async def reset_net_cpu(self):
        """Reset the NET CPU."""
        self.platform.debug_log('Resetting NET CPU.')
        self.writer.write('BR:\r'.encode())
        msg = ''
        while msg != 'BR:P\r' and not msg.endswith('!B:02\r'):
            msg = (await self.readuntil(b'\r')).decode()
            self.platform.debug_log("Got: {}".format(msg))

    async def query_fast_io_boards(self):
        """Query the NET processor to see if any FAST IO boards are connected.

        If so, queries the IO boards to log them and make sure they're the  proper firmware version.
        """
        # reset CPU early
        try:
            await asyncio.wait_for(self.reset_net_cpu(), 5)
        except asyncio.TimeoutError:
            self.platform.warning_log("Reset of NET CPU failed. This might be a firmware bug in your version.")
        else:
            self.platform.debug_log("Reset successful")

        await asyncio.sleep(.5)

        self.platform.debug_log('Reading all switches.')
        self.writer.write('SA:\r'.encode())
        msg = ''
        while not msg.startswith('SA:'):
            msg = (await self.readuntil(b'\r')).decode()
            if not msg.startswith('SA:'):
                self.platform.log.warning("Got unexpected message from FAST: {}".format(msg))

        self.platform.process_received_message(msg, "NET")
        self.platform.debug_log('Querying FAST IO boards...')

        firmware_ok = True

        for board_id in range(128):
            self.writer.write('NN:{:02X}\r'.format(board_id).encode())
            msg = ''
            while not msg.startswith('NN:'):
                msg = (await self.readuntil(b'\r')).decode()
                if not msg.startswith('NN:'):
                    self.platform.debug_log("Got unexpected message from FAST: {}".format(msg))

            if msg == 'NN:F\r':
                break

            node_id, model, fw, dr, sw, _, _, _, _, _, _ = msg.split(',')
            node_id = node_id[3:]

            model = model.strip('\x00')

            # Iterate as many boards as possible
            if not model or model == '!Node Not Found!':
                break

            self.platform.register_io_board(FastIoBoard(int(node_id, 16), model, fw, int(sw, 16), int(dr, 16)))

            self.platform.debug_log('Fast IO Board {0}: Model: {1}, '
                                    'Firmware: {2}, Switches: {3}, '
                                    'Drivers: {4}'.format(node_id,
                                                          model, fw,
                                                          int(sw, 16),
                                                          int(dr, 16)))

            if StrictVersion(IO_MIN_FW) > str(fw):
                self.platform.log.critical("Firmware version mismatch. MPF "
                                           "requires the IO boards to be firmware {0}, but "
                                           "your Board {1} ({2}) is v{3}".format(IO_MIN_FW, node_id, model, fw))
                firmware_ok = False

        if not firmware_ok:
            raise AssertionError("Exiting due to IO board firmware mismatch")

    def send(self, msg):
        """Send a message to the remote processor over the serial connection.

        Args:
        ----
            msg: String of the message you want to send. THe <CR> character will
                be added automatically.

        """
        self.send_queue.put_nowait(msg)

    def _send(self, msg):
        debug = self.platform.config['debug']
        if self.dmd:
            self.writer.write(b'BM:' + msg)
            if debug:
                self.platform.log.debug("Send: %s", "".join(" 0x%02x" % b for b in msg))

        else:
            self.messages_in_flight += 1
            if self.messages_in_flight > self.max_messages_in_flight:
                self.send_ready.clear()

                self.log.debug("Enabling Flow Control for %s connection. "
                               "Messages in flight: %s, Max setting: %s",
                               self.remote_processor,
                               self.messages_in_flight,
                               self.max_messages_in_flight)

            self.writer.write(msg.encode() + b'\r')
            if debug and msg[0:2] != "WD":
                self.platform.log.debug("Send: %s", msg)

    async def _socket_writer(self):
        while True:
            msg = await self.send_queue.get()
            try:
                await asyncio.wait_for(self.send_ready.wait(), 1.0)
            except asyncio.TimeoutError:
                self.log.warning("Port %s was blocked for more than 1s. Reseting send queue! If this happens "
                                 "frequently report a bug!", self.port)
                self.messages_in_flight = 0
                self.send_ready.set()

            self._send(msg)

    def __repr__(self):
        """Return str representation."""
        return "{} ({})".format(self.port, self.remote_processor)

    def _parse_msg(self, msg):
        self.received_msg += msg

        while True:
            pos = self.received_msg.find(b'\r')

            # no more complete messages
            if pos == -1:
                break

            msg = self.received_msg[:pos]
            self.received_msg = self.received_msg[pos + 1:]

            if msg[:2] not in self.ignored_messages_in_flight:

                self.messages_in_flight -= 1
                if self.messages_in_flight <= self.max_messages_in_flight or not self.read_task:
                    self.send_ready.set()
                if self.messages_in_flight < 0:
                    self.log.warning("Port %s received more messages than "
                                     "were sent! Resetting!",
                                     self.remote_processor)
                    self.messages_in_flight = 0

            if not msg:
                continue

            if msg.decode() not in self.ignored_messages:
                self.platform.process_received_message(msg.decode(), self.remote_processor)



import time

from functools import partial
from mpf.core.logging import LogMixin

from mpf.tests.MpfGameTestCase import MpfGameTestCase


class BenchmarkLightShows(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'benchmarks/machine_files/shows/'

    def get_options(self):
        options = super().get_options()
        if self.unittest_verbosity() <= 1:
            options["production"] = True
        return options

    def get_platform(self):
        return 'virtual'

    def setUp(self):
        LogMixin.unit_test = False
        super().setUp()

    def _output(self, name, start, end, end2, num):
        print("Duration {} {:.5f}ms Processing {:.5f}ms Total: {:5f}ms Per second: {:2f}".format(
            name,
            (1000 * (end - start) / num), ((end2 - end) * 1000) / num, (1000 * (end2 - start)) / num,
            (num * 1) / (end2 - start)
        ))

    def _benchmark(self, function, name, num=10000, iterations=10):
        function(num, True)
        total = 0
        for i in range(iterations):
            start, end, end2 = function(num, False)
            total += (end2 - start) / num
            self._output(name, start, end, end2, num)

        print("Total average {:.5f}ms".format(total * 1000/ iterations))
        return total/iterations

    def testBenchmark(self):
        baseline = self._benchmark(partial(self._event_and_run, "random_event", "random_event2"), "baseline")
        minimal_show = self._benchmark(partial(self._event_and_run, "play_minimal_light_show", "stop_minimal_light_show"), "minimal_show")
        minimal_show_token = self._benchmark(partial(self._event_and_run, "play_minimal_light_show_token", "stop_minimal_light_show_token"), "minimal_show_token")
        all_leds = self._benchmark(partial(self._event_and_run, "play_single_step_tag_playfield", "stop_single_step_tag_playfield"), "all_leds_tag")
        multi_step = self._benchmark(partial(self._event_and_run, "play_multi_step", "stop_multi_step"), "multi_step", num=500)

        print("Baseline: {:.5f}ms One LED: +{:.5f}ms {:.2f} fps; "
              "One LED (token): +{:.5f}ms {:.2f} fps; "
              "30 LEDs: +{:.5f}ms {:.2f} fps; "
              "Multi Step: +{:.5f} {:.2f} fps;".format(
            baseline * 1000,
            (minimal_show - baseline) * 1000, 1 / (minimal_show - baseline),
            (minimal_show_token - baseline) * 1000, 1 / (minimal_show_token - baseline),
            (all_leds - baseline) * 1000, 1 / (all_leds - baseline),
            (multi_step - baseline) * 1000, 1 / (multi_step - baseline)
            ))

    def _event_and_run(self, event, event2, num, test):
        channel_list = []
        for light in self.machine.lights.values():
            for color, channels in light.hw_drivers.items():
                channel_list.extend(channels)

        start = time.time()
        for i in range(num):
            self.machine.events.post(event)
            for channel in channel_list:
                brightness = channel.current_brightness
            self.advance_time_and_run(.01)
            for channel in channel_list:
                brightness = channel.current_brightness
            self.machine.events.post(event2)

        end = time.time()
        self.advance_time_and_run()
        end2 = time.time()
        self.machine.events.post(event2)
        return start, end, end2



"""Yaml interface in roundtrip mode."""
from typing import Any

import ruamel.yaml
import ruamel.yaml.emitter
import ruamel.yaml.serializer
import ruamel.yaml.representer

from ruamel.yaml import StringIO, YAML, VersionedResolver
from ruamel.yaml.reader import Reader
from ruamel.yaml.scanner import RoundTripScanner
from ruamel.yaml.parser import Parser
from ruamel.yaml.composer import Composer
from ruamel.yaml.constructor import RoundTripConstructor
from ruamel.yaml.dumper import RoundTripDumper

from mpf.file_interfaces.yaml_interface import YamlInterface, MpfResolver, RESOLVERS

# pylint: disable-msg=invalid-name
typ = 'mpf-rt'


class MpfRoundTripResolver(VersionedResolver):

    """Resolver with mentioned fixes."""


class FormattedInt(ruamel.yaml.scalarint.ScalarInt):

    """An integer which preserves the formatting."""

    def __new__(cls, *args, **kw):
        """Preserve raw string representation."""
        # pylint: disable-msg=protected-access
        x = ruamel.yaml.scalarint.ScalarInt.__new__(cls, *args, **kw)
        x.raw = args[0]
        return x


def alt_construct_yaml_int(constructor, node):
    """Parse integers using FormattedInt."""
    value_s = ruamel.yaml.compat.to_str(constructor.construct_scalar(node))
    if value_s.isdigit():
        return constructor.construct_yaml_int(node)
    return FormattedInt(value_s)


ruamel.yaml.constructor.RoundTripConstructor.add_constructor(
    u'tag:yaml.org,2002:int', alt_construct_yaml_int)


def represent_int(representer, data):
    """Return raw representation for FormattedInt."""
    return representer.represent_scalar(u'tag:yaml.org,2002:int',
                                        str(data.raw))


ruamel.yaml.representer.RoundTripRepresenter.add_representer(FormattedInt, represent_int)


for resolver in RESOLVERS:
    MpfRoundTripResolver.add_implicit_resolver(*resolver)


def init_typ(self):
    """Initialise ruamel RoundTrip."""
    self.Resolver = MpfRoundTripResolver
    self.default_flow_style = False
    # no optimized rt-dumper yet
    self.Emitter = ruamel.yaml.emitter.Emitter  # type: Any
    self.Serializer = ruamel.yaml.serializer.Serializer  # type: Any
    self.Representer = ruamel.yaml.representer.RoundTripRepresenter  # type: Any
    self.Scanner = ruamel.yaml.scanner.RoundTripScanner  # type: Any
    # no optimized rt-parser yet
    self.Parser = ruamel.yaml.parser.RoundTripParser  # type: Any
    self.Composer = ruamel.yaml.composer.Composer  # type: Any
    self.Constructor = ruamel.yaml.constructor.RoundTripConstructor  # type: Any


class MpfRoundTripLoader(Reader, RoundTripScanner, Parser, Composer, RoundTripConstructor, MpfResolver):

    """Config loader which can roundtrip."""

    def __init__(self, stream, version=None, preserve_quotes=None):
        """Initialise loader."""
        del version
        Reader.__init__(self, stream, loader=self)
        RoundTripScanner.__init__(self, loader=self)
        Parser.__init__(self, loader=self)
        Composer.__init__(self, loader=self)
        RoundTripConstructor.__init__(self, preserve_quotes=preserve_quotes, loader=self)
        MpfResolver.__init__(self, loadumper=self)


class YamlRoundtrip(YamlInterface):     # pragma: no cover

    """Round trip yaml interface used for the migrator."""

    @staticmethod
    def process(data_string):
        """Parse yaml from a string."""
        return ruamel.yaml.load(data_string, Loader=MpfRoundTripLoader)

    @staticmethod
    def save_to_str(data):
        """Return yaml string from config."""
        return ruamel.yaml.dump(data, Dumper=RoundTripDumper,
                                default_flow_style=False, indent=4, width=10)

    def save(self, filename, data):
        """Save config to yaml file."""
        with open(filename, 'w', encoding='utf8') as output_file:
            output_file.write(ruamel.yaml.dump(data, Dumper=RoundTripDumper))

    @staticmethod
    def rename_key(old_key, new_key, commented_map, logger=None):
        """Rename a key in YAML file data that was loaded with the RoundTripLoader (e.g. that contains comments).

        Comments are retained for the renamed key. Order of keys is also maintained.

        Args:
        ----
            old_key: The existing key name you want to change.
            new_key: The new key name.
            commented_map: The YAML data CommentMap class (from yaml.load) with
                the key you want to change.
            logger: Optional logger instance which will be used to log this at
                the debug level.

        Returns the updated CommentedMap YAML dict. (Note that this method does not
        change the dict object (e.g. it's changed in place), you you most
        likely don't need to do anything with the returned dict.
        """
        if old_key == new_key or old_key not in commented_map:
            return commented_map

        key_list = list(commented_map.keys())

        for key in key_list:
            if key == old_key:

                if logger:
                    logger.debug('Renaming key: %s: -> %s:', old_key, new_key)

                commented_map[new_key] = commented_map[old_key]

                try:  # if there's no comment, it will not be in ca.items
                    commented_map.ca.items[new_key] = (
                        commented_map.ca.items.pop(old_key))
                except KeyError:
                    pass

                del commented_map[old_key]

            else:
                commented_map.move_to_end(key)

        return commented_map

    # pylint: disable-msg=too-many-arguments
    @staticmethod
    def copy_with_comments(source_dict, source_key, dest_dict, dest_key, delete_source=False, logger=None):
        """Copy dict with comments in config."""
        dest_dict[dest_key] = source_dict[source_key]

        try:
            dest_dict.ca.items[dest_key] = source_dict.ca.items[source_key]

            if logger and not delete_source:
                logger.debug('Copying key: %s -> %s', source_key, dest_key)

        except (KeyError, AttributeError):
            pass

        if delete_source:

            if logger:
                logger.debug('Moving key: %s -> %s', source_key, dest_key)

            del source_dict[source_key]
            source_dict.ca.items.pop(source_key, None)

    @staticmethod
    def del_key_with_comments(dic, key, logger=None):
        """Delete section with comments."""
        if key not in dic:
            return

        if logger:
            logger.debug("Removing key: %s", key)

        del dic[key]
        dic.ca.items.pop(key, None)

    @staticmethod
    def pretty_format(dic):
        """Return pretty printed config."""
        return '\r' + ruamel.yaml.dump(dic, Dumper=RoundTripDumper, indent=4)

    @classmethod
    def reformat_file(cls, filename, show_file=False) -> bool:
        """Reformat a yaml config file.

        Returns true if the file was changed.
        """
        with open(filename) as f:
            content = f.read()

        formatted_yaml_string = cls.reformat_yaml(content, show_file=show_file)
        if content != formatted_yaml_string:
            with open(filename, "w") as f:
                f.write(formatted_yaml_string)
            return True

        # nothing changed
        return False

    @staticmethod
    def reformat_yaml(yaml_string, show_file=False):
        """Reformat a yaml config string."""
        yaml_obj = YAML(typ="mpf-rt", plug_ins=["mpf.file_interfaces.yaml_roundtrip"])
        if show_file:
            yaml_obj.indent(mapping=2, sequence=2, offset=0)
        else:
            yaml_obj.indent(mapping=2, sequence=4, offset=2)

        yaml_obj.preserve_quotes = True
        yaml_obj.width = 10000
        data = yaml_obj.load(yaml_string)
        string_stream = StringIO()
        yaml_obj.dump(data, string_stream)
        formatted_yaml_string = string_stream.getvalue()
        # if show_file:
        #    formatted_yaml_string = re.sub(r'^  ', '', formatted_yaml_string, flags=re.MULTILINE)

        return formatted_yaml_string



"""Contains the drivers and interface code for pinball machines which use the Multimorphic P3-ROC hardware controllers.

Much of this code is from the P-ROC drivers section of the pyprocgame project,
written by Adam Preble and Gerry Stellenberg. It was originally released under
the MIT license and is released here under the MIT License.

More info on the P3-ROC hardware platform: http://pinballcontrollers.com/

Original code source on which this module was based:
https://github.com/preble/pyprocgame
"""
import asyncio
import logging

from typing import Dict, List, Optional  # pylint: disable-msg=cyclic-import,unused-import

from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface

from mpf.platforms.interfaces.i2c_platform_interface import I2cPlatformInterface

from mpf.core.platform import I2cPlatform, AccelerometerPlatform, DriverConfig, SwitchConfig
from mpf.platforms.interfaces.accelerometer_platform_interface import AccelerometerPlatformInterface
from mpf.platforms.p_roc_common import PDBConfig, PROCBasePlatform
from mpf.platforms.p_roc_devices import PROCDriver

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.accelerometer import Accelerometer     # pylint: disable-msg=cyclic-import,unused-import


class P3RocGpioSwitch(SwitchPlatformInterface):

    """P3-ROC switch on GPIOs."""

    __slots__ = ["index"]

    def __init__(self, config, number, index):
        """Initialise P-ROC switch."""
        super().__init__(config, number)
        self.index = index

    def get_board_name(self):
        """Return board of the GPIOs."""
        return "P3-Roc GPIOs"

    @property
    def has_rules(self):
        """Return false as we do not support rules."""
        return False


class P3RocGpioDriver(DriverPlatformInterface):

    """P3-Roc driver in GPIOs."""

    __slots__ = ["index", "platform"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, number, config, index, platform):
        """Initialise driver."""
        super().__init__(config, number)
        self.index = index
        self.platform = platform

    def get_board_name(self):
        """Return board of the GPIOs."""
        return "P3-Roc GPIOs"

    def disable(self):
        """Disable (turn off) this GPIO."""
        self.platform.set_gpio(self.index, False)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable (turn on) this GPIO."""
        if pulse_settings.power != 1 or hold_settings.power != 1:
            raise AssertionError("pulse_power and hold_power both must be 1.0 for GPIO drivers. Let us know in the "
                                 "forum if you need this.")

        self.platform.set_gpio(self.index, True)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse GPIO."""
        raise AssertionError("Not currently implemented. Let us know in the forum if you need pulses on GPIOs.")

    @property
    def has_rules(self):
        """Return false as we do not support rules."""
        return False


class P3RocHardwarePlatform(PROCBasePlatform, I2cPlatform, AccelerometerPlatform):

    """Platform class for the P3-ROC hardware controller.

    Args:
    ----
        machine: The MachineController instance.
    """

    __slots__ = ["_burst_opto_drivers_to_switch_map", "_burst_switches", "_bursts_enabled", "acceleration",
                 "accelerometer_device", "gpio_poll_task", "gpio_config"]

    def __init__(self, machine):
        """Initialise and connect P3-Roc."""
        super().__init__(machine)
        # validate config for p3_roc
        self.config = self.machine.config_validator.validate_config("p3_roc", self.machine.config.get('p_roc', {}))
        self._configure_device_logging_and_debug('P-Roc', self.config)

        if self.config['driverboards']:
            self.machine_type = self.pinproc.normalize_machine_type(self.config['driverboards'])
        else:
            self.machine_type = self.pinproc.normalize_machine_type(self.machine.config['hardware']['driverboards'])

        self.debug = self.config["debug"]

        if self.machine_type != self.pinproc.MachineTypePDB:
            raise AssertionError("P3-Roc can only handle PDB driver boards")

        # Because PDBs can be configured in many different ways, we need to
        # traverse the YAML settings to see how many PDBs are being used.
        # Then we can configure the P3-ROC appropriately to use those PDBs.
        # Only then can we relate the YAML coil/light #'s to P3-ROC numbers for
        # the collections.

        self.debug_log("Configuring P3-ROC for PDB driver boards.")
        self._burst_opto_drivers_to_switch_map = {}
        self._burst_switches = []   # type: List[P3RocBurstOpto]
        self._bursts_enabled = False
        self.gpio_poll_task = None
        self.gpio_config = 0

        self.acceleration = [0] * 3
        self.accelerometer_device = None    # type: Optional[PROCAccelerometer]

    async def connect(self):
        """Connect to the P3-Roc."""
        await super().connect()

        self.pdbconfig = PDBConfig(self, self.machine.config, self.pinproc.DriverCount)

        if self.dipswitches & 0x01:
            self.log.info("Burst drivers are configured as outputs (DIP Switch 1 set). "
                          "You cannot use IDs 0-3 for PD-16/PD-LED boards.")

            if self.version < 2 or self.revision < 6:
                raise AssertionError("Local inputs are supported only in FW 2.6+. Disable DIP 1 or update firmware.")

        if self.dipswitches & 0x02:
            self.log.info("Burst switches are configured as inputs (DIP Switch 2 set). "
                          "You cannot use IDs 0-3 for SW-16 boards.")

            if self.version < 2 or self.revision < 6:
                raise AssertionError("Local inputs are supported only in FW 2.6+. Disable DIP 2 or update firmware.")

            for board in range(0, 4):
                device_type = await self.run_proc_cmd("read_data", 2, (1 << 12) + (board << 6))
                if device_type != 0:
                    raise AssertionError("Invalid P3-Roc configuration. Found SW-16 with ID {} which is invalid "
                                         "because burst switches/drivers which are configured as inputs/outputs use "
                                         "the same switch position. Either disabled DIP 2 or assign ID >= 4 to "
                                         "all your SW-16s.")

        # remove all burst ir mappings
        for driver in range(0, 64):
            self.run_proc_cmd_no_wait("write_data", 0x02, 0x80 + (driver * 2), 0)
            self.run_proc_cmd_no_wait("write_data", 0x02, 0x81 + (driver * 2), 0)

        # disable burst IRs
        burst_config1 = 0
        self.run_proc_cmd_no_wait("write_data", 0x02, 0x01, burst_config1)

    async def start(self):
        """Start GPIO poller."""
        await super().start()
        if self.config["gpio_map"]:
            has_inputs = False
            for gpio_index, state in self.config["gpio_map"].items():
                if state == "output":
                    self.gpio_config += 1 << (gpio_index + 8)
                else:
                    has_inputs = True

            if has_inputs:
                self.gpio_poll_task = self.machine.clock.loop.create_task(self._poll_gpios())
                self.gpio_poll_task.add_done_callback(Util.raise_exceptions)

            self.run_proc_cmd_no_wait("write_data", 0x00, 0x03, self.gpio_config)

    async def _poll_gpios(self):
        """Poll GPIOs."""
        poll_sleep = 1 / self.config["gpio_poll_frequency"]
        gpio_state_old = None
        while True:
            await asyncio.sleep(poll_sleep)
            gpio_state = await self.run_proc_cmd("read_data", 0x00, 0x04)
            for gpio_index, state in self.config["gpio_map"].items():
                if state == "input" and (gpio_state_old is None or (gpio_state ^ gpio_state_old) & (1 << gpio_index)):
                    self.machine.switch_controller.process_switch_by_num("gpio-{}".format(gpio_index),
                                                                         bool(gpio_state & (1 << gpio_index)), self)
            gpio_state_old = gpio_state

    def set_gpio(self, index, state):
        """Set GPIO state."""
        new_gpio_config = self.gpio_config
        if state:
            new_gpio_config |= 1 << index
        else:
            new_gpio_config &= 0xffff ^ (1 << index)

        if new_gpio_config != self.gpio_config:
            self.gpio_config = new_gpio_config
            self.run_proc_cmd_no_wait("write_data", 0x00, 0x03, self.gpio_config)

    def stop(self):
        """Stop platform."""
        super().stop()
        if self.gpio_poll_task:
            self.gpio_poll_task.cancel()
            self.gpio_poll_task = None

    def _get_default_subtype(self):
        """Return default subtype for P3-Roc."""
        return "led"

    def __repr__(self):
        """Return string representation."""
        return '<Platform.P3-ROC>'

    async def configure_i2c(self, number: str):
        """Configure I2C device on P3-Roc."""
        return P3RocI2c(number, self)

    @classmethod
    def scale_accelerometer_to_g(cls, raw_value):
        """Convert internal representation to g."""
        # raw value is 0 to 16384 -> 14 bit
        # scale is -2g to 2g (2 complement)
        if raw_value & (1 << 13):
            raw_value -= 1 << 14

        g_value = float(raw_value) / (1 << 12)

        return g_value

    def configure_accelerometer(self, number, config, callback):
        """Configure the accelerometer on the P3-ROC."""
        del config
        if number != "1":
            raise AssertionError("P3-ROC only has one accelerometer. Use number 1. Found: {}".format(number))

        self.accelerometer_device = PROCAccelerometer(callback)
        self._configure_accelerometer()

        return self.accelerometer_device

    def _configure_accelerometer(self):

        # enable polling every 128ms
        enable = 0x0F

        # configure some P3-Roc registers

        self.run_proc_cmd_no_wait("write_data", 6, 0x000, enable)

        # CTRL_REG1 - set to standby
        self.run_proc_cmd_no_wait("write_data", 6, 0x12A, 0)

        # XYZ_DATA_CFG - disable high pass filter, scale 0 to 2g
        self.run_proc_cmd_no_wait("write_data", 6, 0x10E, 0x00)

        # CTRL_REG1 - set device to active and in low noise mode
        # 800HZ output data rate
        self.run_proc_cmd_no_wait("write_data", 6, 0x12A, 0x05)

        # CTRL_REG2 - set no sleep, high resolution mode
        self.run_proc_cmd_no_wait("write_data", 6, 0x12B, 0x02)

        # for auto-polling of accelerometer every 128 ms (8 times a sec). set 0x0F
        # disable polling + IRQ status addr FF_MT_SRC
        self.run_proc_cmd_no_wait("write_data", 6, 0x000, 0x1E0F)
        # flush data to proc
        self.run_proc_cmd_no_wait("flush")

    def get_info_string(self):
        """Dump infos about boards."""
        infos = "Firmware Version: {} Firmware Revision: {} Hardware Board ID: {}\n".format(
            self.version, self.revision, self.hardware_version)

        infos += "SW-16 boards found:\n"

        for board in range(0, 32):
            device_type = self.run_proc_cmd_sync("read_data", 2, (1 << 12) + (board << 6))
            board_id = self.run_proc_cmd_sync("read_data", 2, (1 << 12) + (board << 6) + 1)
            if device_type != 0:
                infos += " - Board: {} Switches: 16 Device Type: {:X} Board ID: {:X}\n".format(
                    board, device_type, board_id)

        return infos

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Create a P3-ROC driver.

        Typically drivers are coils or flashers, but for the P3-ROC this is
        also used for matrix-based lights.

        Args:
        ----
            config: Dictionary of settings for the driver.
            number: Number of this driver.
            platform_settings: Platform specific settings

        Returns a reference to the PROCDriver object which is the actual object you
        can use to pulse(), patter(), enable(), etc.
        """
        # todo need to add virtual driver support for driver counts > 256

        # Find the P3-ROC number for each driver. For P3-ROC driver boards, the
        # P3-ROC number is specified via the Ax-By-C format.

        if number.startswith("direct-"):
            return self._configure_direct_driver(config, number)
        if number.startswith("gpio-"):
            return self._configure_gpio_driver(config, number)

        proc_num = self.pdbconfig.get_proc_coil_number(str(number))
        if proc_num == -1:
            raise AssertionError("Driver {} cannot be controlled by the P3-ROC. ".format(str(number)))

        if proc_num < 32 and self.dipswitches & 0x01:
            raise AssertionError("Cannot use PD-16 with ID 0 or 1 when DIP 1 is on the P3-Roc. Turn DIP 1 off or "
                                 "renumber PD-16s. Driver: {}".format(number))

        proc_driver_object = PROCDriver(proc_num, config, self, number, True)

        return proc_driver_object

    def _configure_direct_driver(self, config, number):
        try:
            _, driver_number = number.split("-", 2)
            driver_number = int(driver_number)
        except (ValueError, TypeError):
            raise AssertionError("Except format direct-X with 0 <= X <= 63. Invalid format. Got: {}".format(number))

        if 0 < driver_number > 63:
            raise AssertionError("Except format direct-X with 0 <= X <= 63. X out of bounds. Got: {}".format(number))

        if not self.dipswitches & 0x01:
            raise AssertionError("Set DIP 1 on the P3-Roc to use burst switches as local outputs")

        return PROCDriver(driver_number, config, self, number, True)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure a P3-ROC switch.

        Args:
        ----
            number: Number of this switch
            config: Dictionary of settings for the switch.
            platform_config: Platform specific settings.

        Returns: A configured switch object.
        """
        del platform_config
        if number.startswith("burst-"):
            return self._configure_burst_opto(config, number)
        if number.startswith("direct-"):
            return self._configure_direct_switch(config, number)
        if number.startswith("gpio-"):
            return self._configure_gpio_switch(config, number)

        proc_num = self.pdbconfig.get_proc_switch_number(str(number))
        if 0 <= proc_num < 64 and self.dipswitches & 0x02:
            raise AssertionError("Cannot use SW-16 with ID 0-3 when DIP 2 is on the P3-Roc. Turn DIP 2 off or "
                                 "renumber SW-16s. Switch: {}".format(number))
        return self._configure_switch(config, proc_num)

    def _configure_gpio_switch(self, config, number):
        _, switch_number_str = number.split("-", 2)
        index = int(switch_number_str)

        if self.config["gpio_map"].get(index, None) != "input":
            self.raise_config_error("GPIO {} is not configured as input in gpio_map.".format(number), 1)

        return P3RocGpioSwitch(config, number, index)

    def _configure_gpio_driver(self, config, number):
        _, driver_number_str = number.split("-", 2)
        index = int(driver_number_str)

        if self.config["gpio_map"].get(index, None) != "output":
            self.raise_config_error("GPIO {} is not configured as output in gpio_map.".format(number), 2)

        return P3RocGpioDriver(config, number, index, self)

    def _configure_direct_switch(self, config, number):
        try:
            _, switch_number = number.split("-", 2)
            switch_number = int(switch_number)
        except (ValueError, TypeError):
            raise AssertionError("Except format direct-X with 0 <= X <= 63. Invalid format. Got: {}".format(number))

        if 0 < switch_number > 63:
            raise AssertionError("Except format direct-X with 0 <= X <= 63. X out of bounds. Got: {}".format(number))

        if not self.dipswitches & 0x02:
            raise AssertionError("Set DIP 2 on the P3-Roc to use burst switches as local inputs")

        return self._configure_switch(config, switch_number)

    # pylint: disable-msg=too-many-locals
    def _configure_burst_opto(self, config, number):
        """Configure burst opto on the P3-Roc.

        From Gerry:
        Iterate txIndex from 0 to 63 and fill in the rxMap for each.
        The rx map can overlap. Doesn't matter. The P3-ROC hardware drives burst tx0 then checks the 5 mapped rx.
        Then it drives burst tx1 and checks its 5 mapped rx. Etc up to tx63.
        Also be sure to set the Max burst tx in Switch Controller Burst Configuration 1 register (as documented in the
        programmer's reference).
        Only other thing I can think that matters is this. There are 64 burst tx (outputs) on the board, but only 32
        actual drivers in the code. 0:31 are duplicated as 32:63. So at any one time, 2 tx pins are active.

        Now for burst rx map configuration:
        // Program Count into 1st address for transmitter
        data = (uint)rx_list.Count;
        Machine.PROC.write_data (switchModule, (uint)(burstConfigOffset + tx*2), data);

        // Prepare rx map
        data = 0;
        foreach (int receiver in rx_list) {
          data = (data << 6) | ((uint)receiver & 0x3f);
        }
        Machine.PROC.write_data (switchModule, (uint)(burstConfigOffset + tx*2+1), data);
        """
        # parse input and driver switches first
        try:
            _, input_switch, driver = number.split("-", 3)
            input_switch = int(input_switch)
            driver = int(driver)
        except ValueError:
            raise AssertionError("Burst Opto {} is invalid. Format should be burst-XX-YY with X=input Y=driver.")

        # verify we are not conflicting with local inputs
        if self.dipswitches & 0x03:
            raise AssertionError("Cannot use Burst Optos when local inputs or outputs are used. Disable DIP 1 and 2 "
                                 "on the P3-Roc.")

        # enable burst IRs
        if not self._bursts_enabled:
            self._bursts_enabled = True
            self.log.info("Enabling all burst opto on the P3-Roc.")
            burst_config0 = self.config['burst_us_per_half_pulse'] & 0x3F
            burst_config0 |= (self.config['burst_number_of_pulses_to_drive_output'] & 0x1F) << 6
            burst_config0 |= (self.config['burst_number_of_idle_pulses_before_next'] & 0x3F) << 12
            burst_config0 |= (self.config['burst_number_of_burst_pulses_before_check'] & 0x3F) << 18
            burst_config0 |= ((self.config['burst_ms_between_scans'] - 1) & 0x1F) << 24
            self.proc.write_data(0x02, 0x00, burst_config0)
            self.debug_log("Setting 0x02 0x00 to %s", burst_config0)

            burst_config1 = (1 << 31) | 0x1F

            self.run_proc_cmd_no_wait("write_data", 0x02, 0x01, burst_config1)
            self.debug_log("Setting 0x02 0x01 to %s", burst_config1)

            # enable receiver 63 for all of the optos (works around bug in fpga)
            for driver_num in range(0, 64):
                self.run_proc_cmd_no_wait("write_data", 0x02, 0x80 + (driver_num * 2), 1)
                self.debug_log("Setting 0x02 %s to %s", 0x80 + (driver_num * 2), 1)
                self.run_proc_cmd_no_wait("write_data", 0x02, 0x81 + (driver_num * 2), 63)
                self.debug_log("Setting 0x02 %s to %s", 0x81 + (driver_num * 2), 63)

        # configure driver for receiver
        if driver not in self._burst_opto_drivers_to_switch_map:
            self._burst_opto_drivers_to_switch_map[driver] = []

        if input_switch in self._burst_opto_drivers_to_switch_map[driver]:
            raise AssertionError("Input {} already configured for driver {} in {}. Make sure to configure each "
                                 "burst input<->driver combination only once.".format(input_switch, driver, number))

        # tell p3-roc to check this input for that driver
        self._burst_opto_drivers_to_switch_map[driver].append(input_switch)

        if len(self._burst_opto_drivers_to_switch_map[driver]) > 5:
            raise AssertionError("Every burst driver only supports up to 5 drivers. Driver {} exceeded that with "
                                 "switch {}.".format(driver, number))

        rx_to_check_for_this_transmitter = 0
        for switch in self._burst_opto_drivers_to_switch_map[driver]:
            rx_to_check_for_this_transmitter <<= 6
            rx_to_check_for_this_transmitter += switch

        addr_80 = 0x80 + (driver * 2)
        data_80 = len(self._burst_opto_drivers_to_switch_map[driver])
        self.debug_log("Setting 0x02 %s to %s", addr_80, data_80)
        self.run_proc_cmd_no_wait("write_data", 0x02, addr_80, data_80)

        addr_81 = 0x81 + (driver * 2)
        self.debug_log("Setting 0x02 %s to %s", addr_81, rx_to_check_for_this_transmitter)
        self.run_proc_cmd_no_wait("write_data", 0x02, addr_81, rx_to_check_for_this_transmitter)

        burst_switch = P3RocBurstOpto(config, number, input_switch, driver)
        self._burst_switches.append(burst_switch)

        return burst_switch

    async def get_hw_switch_states(self) -> Dict[str, bool]:
        """Read in and set the initial switch state.

        The P-ROC uses the following values for hw switch states:
        1 - closed (debounced)
        2 - open (debounced)
        3 - closed (not debounced)
        4 - open (not debounced)
        """
        states = await self.run_proc_cmd("switch_get_states")
        result = {}     # type: Dict[str, bool]

        for switch, state in enumerate(states):
            # Note: The P3-ROC will return a state of "3" for switches from non-
            # connected SW-16 boards, so that's why we only check for "1" below
            result[switch] = bool(state == 1)

        # assume 0 for all bursts initially
        for burst_switch in self._burst_switches:
            result[burst_switch.number] = False

        # read GPIOs
        if self.config["gpio_map"]:
            gpio_state = await self.run_proc_cmd("read_data", 0x00, 0x04)
            for gpio_index, state in self.config["gpio_map"].items():
                if state == "input":
                    result["gpio-{}".format(gpio_index)] = bool(gpio_state & (1 << gpio_index))

        return result

    def process_events(self, events):
        """Process events from the P3-Roc."""
        for event in events:
            event_type = event['type']
            event_value = event['value']
            if event_type == self.pinproc.EventTypeSwitchClosedDebounced:
                self.machine.switch_controller.process_switch_by_num(state=1,
                                                                     num=event_value,
                                                                     platform=self)
            elif event_type == self.pinproc.EventTypeSwitchOpenDebounced:
                self.machine.switch_controller.process_switch_by_num(state=0,
                                                                     num=event_value,
                                                                     platform=self)
            elif event_type == self.pinproc.EventTypeSwitchClosedNondebounced:
                self.machine.switch_controller.process_switch_by_num(state=1,
                                                                     num=event_value,
                                                                     platform=self)
            elif event_type == self.pinproc.EventTypeSwitchOpenNondebounced:
                self.machine.switch_controller.process_switch_by_num(state=0,
                                                                     num=event_value,
                                                                     platform=self)

            # The P3-ROC will always send all three values sequentially.
            # Therefore, we will trigger after the Z value
            elif event_type == self.pinproc.EventTypeAccelerometerX:
                self.acceleration[0] = event_value
                if self.debug:
                    self.debug_log("Got Accelerometer value X. Value: %s", event_value)
            elif event_type == self.pinproc.EventTypeAccelerometerY:
                self.acceleration[1] = event_value
                if self.debug:
                    self.debug_log("Got Accelerometer value Y. Value: %s", event_value)
            elif event_type == self.pinproc.EventTypeAccelerometerZ:
                self.acceleration[2] = event_value

                # trigger here
                if self.accelerometer_device:
                    self.accelerometer_device.update_acceleration(
                        self.scale_accelerometer_to_g(self.acceleration[0]),
                        self.scale_accelerometer_to_g(self.acceleration[1]),
                        self.scale_accelerometer_to_g(self.acceleration[2]))
                if self.debug:
                    self.debug_log("Got Accelerometer value Z. Value: %s", event_value)
            elif event_type == self.pinproc.EventTypeBurstSwitchOpen:
                if self.debug:
                    self.debug_log("Got burst open event value %s", event_value)
                self._handle_burst(event_value, 0)
            elif event_type == self.pinproc.EventTypeBurstSwitchClosed:
                if self.debug:
                    self.debug_log("Got burst closed event value %s", event_value)
                self._handle_burst(event_value, 1)
            else:   # pragma: no cover
                self.log.warning("Received unrecognized event from the P3-ROC. "
                                 "Type: %s, Value: %s", event_type, event_value)

    def _handle_burst(self, event_value, state):
        input_num = event_value & 0x3F
        output_num = (event_value >> 6) & 0x1F
        burst_number1 = "burst-{}-{}".format(input_num, output_num)
        self.machine.switch_controller.process_switch_by_num(state=state,
                                                             num=burst_number1,
                                                             platform=self)
        burst_number2 = "burst-{}-{}".format(input_num, output_num + 32)
        self.machine.switch_controller.process_switch_by_num(state=state,
                                                             num=burst_number2,
                                                             platform=self)


class P3RocI2c(I2cPlatformInterface):

    """I2c device on a P3-Roc."""

    __slots__ = ["platform", "address", "proc"]

    def __init__(self, number: str, platform) -> None:
        """Initialise I2c device on P3_Roc."""
        super().__init__(number)
        self.platform = platform
        self.address = number

    def i2c_write8(self, register, value):
        """Write an 8-bit value to the I2C bus of the P3-Roc."""
        self.platform.run_proc_cmd_no_wait("write_data", 7, int(self.address) << 9 | register, value)

    async def i2c_read8(self, register):
        """Read an 8-bit value from the I2C bus of the P3-Roc."""
        data = await self.platform.run_proc_cmd("read_data", 7, int(self.address) << 9 | register)
        return data & 0xFF

    async def i2c_read_block(self, register, count):
        """Read block via I2C."""
        result = []
        position = 0
        while position < count:
            if count - position == 1:
                data = await self.i2c_read8(register + position)
                result.append(data)
                position += 1
            else:
                data = await self.i2c_read16(register)
                result.append((data >> 8) & 0xFF)
                result.append(data & 0xFF)
                position += 2
        return result

    async def i2c_read16(self, register) -> int:
        """Read an 16-bit value from the I2C bus of the P3-Roc."""
        return self.platform.run_proc_cmd("read_data", 7, int(self.address) << 9 | 1 << 8 | register)


class PROCAccelerometer(AccelerometerPlatformInterface):

    """The accelerometer on the P3-Roc."""

    __slots__ = ["callback"]

    def __init__(self, callback: "Accelerometer") -> None:
        """Remember the callback."""
        self.callback = callback    # type: Accelerometer

    def update_acceleration(self, x: float, y: float, z: float) -> None:
        """Call the callback."""
        self.callback.update_acceleration(x, y, z)


class P3RocBurstOpto(SwitchPlatformInterface):

    """A burst opto switch/driver combination."""

    __slots__ = ["input_switch", "driver", "log"]

    def __init__(self, config, number, input_switch, driver):
        """Initialise burst opto."""
        super().__init__(config, number)
        self.input_switch = input_switch
        self.driver = driver
        self.log = logging.getLogger('P3RocBurstOpto')

    def get_board_name(self):
        """Return board of the switch."""
        return "P3-Roc Burst Opto Input {} Driver {}".format(self.input_switch, self.driver)



"""OPP solenoid wings."""
import logging
from collections import namedtuple

from typing import Optional, Dict

from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings

from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.platforms.opp.opp import OppHardwarePlatform

SwitchRule = namedtuple("SwitchRule", ["pulse_settings", "hold_settings", "recycle", "can_cancel", "delay_ms"])


class OPPSolenoid(DriverPlatformInterface):

    """Driver of an OPP solenoid card."""

    __slots__ = ["sol_card", "log", "switch_rule", "_switches", "_config_state", "platform_settings", "switches"]

    def __init__(self, sol_card, number):
        """Initialise OPP solenoid driver."""
        super().__init__({}, number)
        self.sol_card = sol_card        # type: OPPSolenoidCard
        self.log = sol_card.log
        self.switch_rule = None         # type: SwitchRule
        self.switches = []
        self._config_state = None
        self.platform_settings = dict()     # type: Dict

    def get_board_name(self):
        """Return OPP chain and addr."""
        return "OPP {} Board {}".format(str(self.sol_card.chain_serial), "0x%02x" % self.sol_card.addr)

    def get_minimum_off_time(self, recycle):
        """Return minimum off factor.

        The hardware applies this factor to pulse_ms to prevent the coil from burning.
        """
        if not recycle:
            return 0
        if self.platform_settings['recycle_factor']:
            if self.platform_settings['recycle_factor'] > 7:
                raise AssertionError("Maximum recycle_factor allowed is 7")
            return self.platform_settings['recycle_factor']

        # default to two times pulse_ms
        return 2

    def _kick_coil(self, sol_int, on):
        mask = 1 << sol_int
        msg = bytearray()
        msg.append(self.sol_card.addr)
        msg.extend(OppRs232Intf.KICK_SOL_CMD)
        if on:
            msg.append((mask >> 8) & 0xff)
            msg.append(mask & 0xff)
        else:
            msg.append(0)
            msg.append(0)
        msg.append((mask >> 8) & 0xff)
        msg.append(mask & 0xff)
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        cmd = bytes(msg)
        if self.sol_card.platform.debug:
            self.log.debug("Triggering solenoid driver: %s", "".join(" 0x%02x" % b for b in cmd))
        self.sol_card.platform.send_to_processor(self.sol_card.chain_serial, cmd)

    def disable(self):
        """Disable (turns off) this driver."""
        _, _, solenoid = self.number.split("-")
        sol_int = int(solenoid)
        if self.sol_card.platform.debug:
            self.log.debug("Disabling solenoid %s", self.number)
        self._kick_coil(sol_int, False)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable (turns on) this driver."""
        self.reconfigure_driver(pulse_settings, hold_settings, False)

        _, _, solenoid = self.number.split("-")
        sol_int = int(solenoid)
        if self.sol_card.platform.debug:
            self.log.debug("Enabling solenoid %s", self.number)
        self._kick_coil(sol_int, True)

        # restore rule if there was one
        self.apply_switch_rule()

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse this driver."""
        # reconfigure driver
        self.reconfigure_driver(pulse_settings, None, False)

        _, _, solenoid = self.number.split("-")
        sol_int = int(solenoid)
        if self.sol_card.platform.debug:
            self.log.debug("Pulsing solenoid %s", self.number)
        self._kick_coil(sol_int, True)

        # restore rule if there was one
        self.apply_switch_rule()

    def remove_switch_rule(self):
        """Remove switch rule."""
        self.switch_rule = None
        self.reconfigure_driver(
            PulseSettings(
                duration=self.config.default_pulse_ms if self.config.default_pulse_ms is not None else 10,
                power=self.config.default_pulse_power),
            HoldSettings(power=self.config.default_hold_power),
            True)

    # pylint: disable-msg=too-many-arguments
    def set_switch_rule(self, pulse_settings: PulseSettings, hold_settings: Optional[HoldSettings], recycle: bool,
                        can_cancel: bool, delay_ms: int):
        """Set and apply a switch rule."""
        new_rule = SwitchRule(pulse_settings, hold_settings, recycle, can_cancel, delay_ms)
        if self.switch_rule and self.switch_rule != new_rule:
            raise AssertionError("Cannot set two rule with different driver settings in opp. Old: {} New: {}".format(
                self.switch_rule, new_rule))
        self.switch_rule = new_rule
        self.apply_switch_rule()

    def apply_switch_rule(self):
        """(Re-)Apply a configured switch rule if there is one."""
        if self.switch_rule:
            self.reconfigure_driver(self.switch_rule.pulse_settings, self.switch_rule.hold_settings,
                                    self.switch_rule.recycle)

    def reconfigure_driver(self, pulse_settings: PulseSettings, hold_settings: Optional[HoldSettings],
                           recycle: bool = False):
        """Reconfigure a driver."""
        new_config_state = (pulse_settings, hold_settings, recycle, bool(self.switch_rule))

        # if config would not change do nothing
        if new_config_state == self._config_state:
            return

        self._config_state = new_config_state

        # If hold is 0, set the auto clear bit
        if not hold_settings or not hold_settings.power:
            cmd = ord(OppRs232Intf.CFG_SOL_AUTO_CLR)
            hold = 0
        else:
            cmd = 0
            hold = int(hold_settings.power * 16)
            if hold >= 16:
                if self.sol_card.platform.min_version[self.sol_card.chain_serial] >= 0x00020000:
                    # set flag for full power
                    cmd += ord(OppRs232Intf.CFG_SOL_ON_OFF)
                    hold = 0
                else:
                    hold = 15
            if self.switch_rule and self.switch_rule.delay_ms:
                raise AssertionError("Cannot use hold and delayed pulses in OPP.")

        minimum_off = self.get_minimum_off_time(recycle)
        _, _, solenoid = self.number.split('-')

        # Before version 0.2.0.0 set solenoid input wasn't available.
        # CFG_SOL_USE_SWITCH was used to enable/disable a solenoid.  This
        # will work as long as switches are added using _add_switch_coil_mapping
        if self.switch_rule:
            if self.sol_card.platform.min_version[self.sol_card.chain_serial] < 0x00020000:
                cmd += ord(OppRs232Intf.CFG_SOL_USE_SWITCH)
            elif str(((int(solenoid) & 0x0c) << 1) | (int(solenoid) & 0x03)) in\
                    [switch.split('-')[2] for switch in self.switches]:
                # If driver is using matching switch set CFG_SOL_USE_SWITCH
                # in case config happens after set switch command
                cmd += ord(OppRs232Intf.CFG_SOL_USE_SWITCH)

            if self.switch_rule.can_cancel:
                cmd += ord(OppRs232Intf.CFG_SOL_CAN_CANCEL)

            if self.switch_rule.delay_ms:
                cmd += ord(OppRs232Intf.CFG_SOL_DLY_KICK)

        pulse_len = pulse_settings.duration

        msg = bytearray()
        msg.append(self.sol_card.addr)
        msg.extend(OppRs232Intf.CFG_IND_SOL_CMD)
        msg.append(int(solenoid))
        msg.append(cmd)
        msg.append(pulse_len)
        if self.switch_rule and self.switch_rule.delay_ms:
            msg.append(self.switch_rule.delay_ms)
        else:
            msg.append(hold + (minimum_off << 4))
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        msg.extend(OppRs232Intf.EOM_CMD)
        final_cmd = bytes(msg)

        if self.sol_card.platform.debug:
            self.log.debug("Writing individual config: %s on %s", "".join(" 0x%02x" % b for b in final_cmd),
                           self.sol_card.chain_serial)
        self.sol_card.platform.send_to_processor(self.sol_card.chain_serial, final_cmd)


class OPPSolenoidCard:

    """OPP solenoid card."""

    __slots__ = ["log", "chain_serial", "addr", "mask", "platform", "state", "card_num"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, chain_serial, addr, mask, sol_dict, platform):
        """Initialise OPP solenoid card."""
        self.log = logging.getLogger('OPPSolenoid {} on {}'.format(addr, chain_serial))
        self.chain_serial = chain_serial
        self.addr = addr
        self.mask = mask
        self.platform = platform    # type: OppHardwarePlatform
        self.state = 0
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))

        self.log.debug("Creating OPP Solenoid at hardware address: 0x%02x", addr)

        for index in range(0, 16):
            if ((1 << index) & mask) != 0:
                number = chain_serial + '-' + self.card_num + '-' + str(index)
                opp_sol = OPPSolenoid(self, number)
                opp_sol.config = {}
                sol_dict[number] = opp_sol
